from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timezone
from numbers import Number
from typing import Optional

from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.datetime_conversions import (
    convert_datetime_to_str,
    convert_str_to_datetime,
)
from soda_core.common.exceptions import ContractParserException
from soda_core.common.logging_constants import Emoticons, ExtraKeys, soda_logger
from soda_core.common.logs import Location
from soda_core.common.yaml import (
    ContractYamlSource,
    VariableResolver,
    YamlList,
    YamlObject,
    YamlValue,
)

logger: logging.Logger = soda_logger

from typing import Protocol


class ContractYamlExtension(Protocol):
    # Extend the contract YAML object. Can modify the state of the contract YAML.
    def extend(self, contract_yaml: "ContractYaml") -> None:
        ...


class ContractYaml:
    """
    Represents YAML as close as possible.
    None means the key was not present.
    If property value types do not match the schema, None value will be in the model
    List properties will have a None value if the property is not present or the content was not a list, a list
    otherwise

    Extensions can add manipulate the contract YAML object.
    Extensions can be registered using the `register_extension` method, and they will be automatically applied.
    """

    contract_yaml_extensions: dict[str, type[ContractYamlExtension]] = {}

    @classmethod
    def register_extension(cls, name: str, extension_cls: type[ContractYamlExtension]) -> None:
        cls.contract_yaml_extensions[name] = extension_cls

    @classmethod
    def parse(
        cls,
        contract_yaml_source: ContractYamlSource,
        provided_variable_values: Optional[dict[str, str]] = None,
        data_timestamp: Optional[str] = None,
        primary_data_source_impl: Optional[DataSourceImpl] = None,
    ) -> Optional[ContractYaml]:
        contract_yaml = ContractYaml(
            contract_yaml_source=contract_yaml_source,
            provided_variable_values=provided_variable_values,
            data_timestamp=data_timestamp,
            primary_data_source_impl=primary_data_source_impl,
        )
        return contract_yaml

    def __init__(
        self,
        contract_yaml_source: ContractYamlSource,
        provided_variable_values: Optional[dict[str, str]],
        data_timestamp: Optional[str] = None,
        primary_data_source_impl: Optional[DataSourceImpl] = None,
    ):
        self.contract_yaml_source: ContractYamlSource = contract_yaml_source
        self.contract_yaml_object: YamlObject = contract_yaml_source.parse()

        self.variables: list[VariableYaml] = self._parse_variable_yamls(contract_yaml_source, provided_variable_values)

        self.execution_timestamp: datetime = datetime.now(timezone.utc)
        self.data_timestamp: datetime = self._get_data_timestamp(data_timestamp, self.execution_timestamp)

        # Some dialects (Dremio) don't use ISO format, we need to override
        f_convert_str_to_datetime = convert_str_to_datetime
        f_convert_datetime_to_str = convert_datetime_to_str
        if primary_data_source_impl:
            f_convert_str_to_datetime = primary_data_source_impl.sql_dialect.convert_str_to_datetime
            f_convert_datetime_to_str = primary_data_source_impl.sql_dialect.convert_datetime_to_str

        soda_variable_values: dict[str, str] = {
            "NOW": f_convert_datetime_to_str(self.execution_timestamp),
            "DATA_TIMESTAMP": f_convert_datetime_to_str(self.data_timestamp),
        }

        self.resolved_variable_values: dict[str, str] = self._resolve_variable_values(
            variable_yamls=self.variables,
            provided_variable_values=provided_variable_values,
            soda_variable_values=soda_variable_values,
        )

        if "NOW" in self.resolved_variable_values:
            now_value = self.resolved_variable_values.get("NOW")
            if f_convert_str_to_datetime(now_value) is None:
                logger.error(f"Variable 'NOW' must be a correct ISO 8601 timestamp format: {now_value}")

        self.contract_yaml_source.resolve_on_read_value(
            resolved_variable_values=self.resolved_variable_values, soda_values=soda_variable_values, use_env_vars=True
        )

        self.dataset = self.contract_yaml_object.read_dataset_identifier("dataset")

        self.check_attributes = self.contract_yaml_object.read_object_opt(
            "check_attributes", default_value={}
        ).to_dict()

        self.filter: Optional[str] = self.contract_yaml_object.read_string_opt("filter")
        if self.filter:
            self.filter = self.filter.strip()

        self.columns: list[ColumnYaml] = self._parse_columns(self.contract_yaml_object)
        self.checks: Optional[list[Optional[CheckYaml]]] = self._parse_checks(self.contract_yaml_object)

        for extension_cls in ContractYaml.contract_yaml_extensions.values():
            try:
                extension = extension_cls()
                extension.extend(self)
            except Exception as e:
                logger.error(
                    f"Error extending contract YAML with extension {extension_cls.__name__}: {e}",
                )

    def _parse_variable_yamls(self, contract_yaml_source, variables) -> list[VariableYaml]:
        variable_yamls: list[VariableYaml] = []

        if self.contract_yaml_object:
            variables_yaml_object: Optional[YamlObject] = self.contract_yaml_object.read_object_opt("variables")
            if variables_yaml_object:
                for variable_name, variable_yaml_object in variables_yaml_object.items():
                    variable_yaml: VariableYaml = VariableYaml(variable_name, variable_yaml_object)
                    variable_yamls.append(variable_yaml)

        return variable_yamls

    def _resolve_variable_values(
        self,
        variable_yamls: list[VariableYaml],
        provided_variable_values: Optional[dict[str, str]],
        soda_variable_values: Optional[dict[str, str]],
    ) -> dict[str, str]:
        variable_values: dict[str, str] = {}

        # Initializing the declared variables
        for variable_yaml in variable_yamls:
            variable_name: str = variable_yaml.name
            variable_values[variable_name] = (
                provided_variable_values.get(variable_name)
                if isinstance(provided_variable_values, dict) and variable_name in provided_variable_values
                else variable_yaml.default
            )

        for variable_yaml in variable_yamls:
            if variable_values.get(variable_yaml.name) is None:
                logger.error(f"Required variable '{variable_yaml.name}' did not get a value")

        if isinstance(provided_variable_values, dict) and "NOW" in provided_variable_values:
            now_str: str = provided_variable_values["NOW"]
            if not isinstance(now_str, str):
                logger.error(f"Provided 'NOW' variable must be a string, but was: {now_str.__class__.__name__}")
            else:
                if convert_str_to_datetime(now_str) is None:
                    logger.error(f"Provided 'NOW' variable value is not a correct ISO 8601 timestamp format: {now_str}")
                variable_values["NOW"] = now_str
        # else:
        #     # Default now initialization
        #     variable_values["NOW"] = convert_datetime_to_str(datetime.now(timezone.utc))

        return self._resolve_variables(variable_values=variable_values, soda_variable_values=soda_variable_values)

    @classmethod
    def _resolve_variables(
        cls, variable_values: Optional[dict[str, str]], soda_variable_values: Optional[dict[str, str]]
    ) -> dict[str, str]:
        """
        Resolve all variables in the dictionary, replacing ${variable_name} expressions
        with their corresponding values, while detecting circular dependencies.

        Args:
            variable_values (dict): Dictionary with string keys and string values
                             containing ${variable_name} expressions

        Returns:
            dict: Dictionary with all variables resolved
        """
        # Create a copy of the input dict to avoid modifying the original
        variable_values = variable_values.copy()

        # Keep track of variables being processed to detect circular references
        processing_stack = set()

        def resolve_value(name: str, value: str) -> str:
            """
            Recursively resolve all variables in a given value.

            Args:
                name (str): The variable name being resolved (for circular detection)
                value (str): The value to resolve

            Returns:
                str: The resolved value
            """
            # Check for circular reference
            if name in processing_stack:
                # Circular reference detected - return the original expression
                # to prevent infinite recursion
                return value

            # Add current variable to the processing stack
            processing_stack.add(name)

            # Replace all variable references in the value
            resolved = VariableResolver.resolve(
                source_text=value,
                variable_values=variable_values,
                soda_variable_values=soda_variable_values,
                use_env_vars=False,
            )

            # Remove current variable from the processing stack
            processing_stack.remove(name)

            return resolved

        # Resolve each variable in the dictionary
        for name in variable_values:
            variable_values[name] = resolve_value(name, variable_values[name])

        for name, value in variable_values.items():
            logger.debug(f"var.{name} = {value}")

        return variable_values

    def _parse_columns(self, contract_yaml_object: YamlObject) -> list[ColumnYaml]:
        columns: Optional[list[Optional[ColumnYaml]]] = None
        if contract_yaml_object:
            column_yaml_objects: Optional[YamlList] = contract_yaml_object.read_list_of_objects_opt("columns")
            if not column_yaml_objects:
                raise ContractParserException(
                    "The contract is missing the required 'columns' property", str(contract_yaml_object.location)
                )
            if isinstance(column_yaml_objects, YamlList):
                columns = []
                column_locations_by_name: dict[str, list[Optional[Location]]] = {}
                for column_yaml_object in column_yaml_objects:
                    if isinstance(column_yaml_object, YamlObject):
                        column_yaml: ColumnYaml = ColumnYaml(contract_yaml=self, column_yaml_object=column_yaml_object)
                        columns.append(column_yaml)
                        if isinstance(column_yaml.name, str):
                            (
                                column_locations_by_name.setdefault(column_yaml.name, []).append(
                                    column_yaml_object.location
                                )
                            )
                for column_name, locations in column_locations_by_name.items():
                    if len(locations) > 1:
                        locations_message: str = ", ".join(
                            [f"[{location.line},{location.column}]" for location in locations if location is not None]
                        )
                        file_location = (
                            f"In {self.contract_yaml_source.file_path} at: "
                            if self.contract_yaml_source.file_path
                            else "At file locations: "
                        )
                        locations_message = f": {file_location}{locations_message}" if locations_message else ""
                        logger.error(f"Duplicate columns with name " f"'{column_name}'{locations_message}")
        return columns

    def _parse_checks(
        self, checks_containing_yaml_object: YamlObject, column_yaml: Optional[ColumnYaml] = None
    ) -> Optional[list[Optional[CheckYaml]]]:
        checks: Optional[list[Optional[CheckYaml]]] = None
        if checks_containing_yaml_object:
            checks_yaml_list: YamlList = checks_containing_yaml_object.read_list_opt("checks")
            if checks_yaml_list:
                if len(list(checks_yaml_list)) == 0:
                    raise ContractParserException(
                        "The 'checks' property must not be an empty list"
                        "Please add at least one check or remove the 'checks' property."
                    )
                checks = []
                for check_index, check_yaml_object in enumerate(checks_yaml_list):
                    check_type_name: Optional[str] = None
                    check_body_yaml_object: Optional[YamlObject] = None

                    if isinstance(check_yaml_object, YamlObject):
                        check_keys: list[str] = check_yaml_object.keys()
                        if len(check_keys) != 1:
                            logging.error(
                                msg=f"Checks require 1 key to be the type",
                                extra={
                                    ExtraKeys.LOCATION: check_yaml_object.location,
                                },
                            )
                        else:
                            check_type_name = check_keys.pop()
                            check_body_yaml_object = check_yaml_object.read_object_opt(key=check_type_name)
                    elif isinstance(check_yaml_object, str):
                        check_type_name = check_yaml_object
                        logger.error(
                            f"{Emoticons.CROSS_MARK} You forgot the " f"colon ':' behind the check '{check_type_name}'."
                        )
                    if isinstance(check_type_name, str):
                        if check_body_yaml_object is None:
                            check_body_yaml_object = YamlObject(
                                yaml_source=checks_containing_yaml_object.yaml_source, yaml_dict={}
                            )
                            check_body_yaml_object.location = checks_yaml_list.create_location_from_yaml_list_index(
                                index=check_index
                            )

                        check_yaml: Optional[CheckYaml] = CheckYaml.parse_check_yaml(
                            check_type_name=check_type_name,
                            check_body_yaml_object=check_body_yaml_object,
                            column_yaml=column_yaml,
                        )
                        if check_yaml:
                            checks.append(check_yaml)
                        else:
                            logger.error(
                                f"Invalid check type '{check_type_name}'. "
                                f"Existing check types: {CheckYaml.get_check_type_names()}"
                            )
                    else:
                        logger.error(f"Checks must have a YAML object structure.")

        return checks

    def _get_data_timestamp(self, data_timestamp: Optional[str], default_soda_now: datetime) -> datetime:
        if isinstance(data_timestamp, str):
            parsed_data_timestamp = convert_str_to_datetime(data_timestamp)
            if isinstance(parsed_data_timestamp, datetime):
                return parsed_data_timestamp
            else:
                logging.error(
                    f"Provided 'data_timestamp' value is not a correct ISO 8601 "
                    f"timestamp format: '{data_timestamp}'"
                )
        return default_soda_now


class VariableYaml:
    def __init__(self, variable_name: str, variable_yaml_object: YamlObject):
        self.variable_yaml_object: YamlObject = variable_yaml_object
        self.name: str = variable_name
        self.type: any = variable_yaml_object.read_string_opt("type") if variable_yaml_object else None
        self.required: any = variable_yaml_object.read_bool_opt("required") if variable_yaml_object else None
        self.default: any = (
            variable_yaml_object.read_value(key="default", expected_type=[str, Number])
            if variable_yaml_object
            else None
        )


class ValidReferenceDataYaml:
    def __init__(self, valid_reference_data_yaml: YamlObject):
        self.dataset: Optional[str] = valid_reference_data_yaml.read_dataset_identifier("dataset")
        self.column: Optional[str] = valid_reference_data_yaml.read_string("column")


@dataclass
class RegexFormat:
    regex: str
    name: Optional[str]

    @classmethod
    def read(cls, yaml_object: YamlObject, key: str) -> Optional[RegexFormat]:
        regex_format_yaml_object: Optional[YamlObject] = yaml_object.read_object_opt(key)
        if regex_format_yaml_object:
            regex: Optional[str] = regex_format_yaml_object.read_string("regex")
            name: Optional[str] = regex_format_yaml_object.read_string("name")
            if isinstance(regex, str):
                return RegexFormat(regex=regex, name=name)
        return None


class MissingAndValidityYaml:
    def __init__(self, yaml_object: YamlObject):
        self.missing_values: Optional[list] = YamlValue.yaml_unwrap(yaml_object.read_list_opt("missing_values"))
        self.missing_format: Optional[RegexFormat] = RegexFormat.read(yaml_object=yaml_object, key="missing_format")

        # cfg_keys = yaml_object.yaml_dict.keys()
        # self.has_missing_configuration_error: bool = ("missing_values" in cfg_keys and self.missing_values is None) or (
        #     "missing_regex_sql" in cfg_keys and self.missing_regex is None
        # )

        self.invalid_values: Optional[list] = YamlValue.yaml_unwrap(yaml_object.read_list_opt("invalid_values"))
        self.invalid_format: Optional[RegexFormat] = RegexFormat.read(yaml_object=yaml_object, key="invalid_format")
        self.valid_values: Optional[list] = YamlValue.yaml_unwrap(yaml_object.read_list_opt("valid_values"))
        self.valid_format: Optional[RegexFormat] = RegexFormat.read(yaml_object=yaml_object, key="valid_format")
        self.valid_min: Optional[Number] = yaml_object.read_number_opt("valid_min")
        self.valid_max: Optional[Number] = yaml_object.read_number_opt("valid_max")
        self.valid_length: Optional[int] = yaml_object.read_number_opt("valid_length")
        self.valid_min_length: Optional[int] = yaml_object.read_number_opt("valid_min_length")
        self.valid_max_length: Optional[int] = yaml_object.read_number_opt("valid_max_length")

        self.valid_reference_data: Optional[ValidReferenceDataYaml] = None
        valid_reference_data_yaml: Optional[YamlObject] = yaml_object.read_object_opt("valid_reference_data")
        if valid_reference_data_yaml:
            self.valid_reference_data = ValidReferenceDataYaml(valid_reference_data_yaml)


class ColumnYaml(MissingAndValidityYaml):
    def __init__(self, contract_yaml: ContractYaml, column_yaml_object: YamlObject):
        self.column_yaml_object: YamlObject = column_yaml_object
        self.name: Optional[str] = column_yaml_object.read_string("name")
        self.data_type: Optional[str] = column_yaml_object.read_string_opt("data_type")
        self.character_maximum_length: Optional[int] = column_yaml_object.read_number_opt("character_maximum_length")
        super().__init__(column_yaml_object)
        self.check_yamls: Optional[list[CheckYaml]] = contract_yaml._parse_checks(
            checks_containing_yaml_object=column_yaml_object, column_yaml=self
        )


class RangeYaml:
    def __init__(self, range_yaml_object: YamlObject):
        self.greater_than: Optional[Number] = range_yaml_object.read_number_opt("greater_than")
        self.greater_than_or_equal: Optional[Number] = range_yaml_object.read_number_opt("greater_than_or_equal")
        self.less_than: Optional[Number] = range_yaml_object.read_number_opt("less_than")
        self.less_than_or_equal: Optional[Number] = range_yaml_object.read_number_opt("less_than_or_equal")

    @classmethod
    def read_range_opt(cls, threshold_yaml_object: YamlObject, key: str) -> Optional[RangeYaml]:
        range_between_yaml_object: Optional[YamlObject] = threshold_yaml_object.read_object_opt(key)
        if range_between_yaml_object:
            return RangeYaml(range_yaml_object=range_between_yaml_object)
        return None

    def __get_common_range_errors(self) -> list[str]:
        errors: list[str] = []
        if isinstance(self.greater_than, Number) and isinstance(self.greater_than_or_equal, Number):
            errors.append("double greater bound")
        if self.greater_than is None and self.greater_than_or_equal is None:
            errors.append("no greater bound")
        if isinstance(self.less_than, Number) and isinstance(self.less_than_or_equal, Number):
            errors.append("double less bound")
        if self.less_than is None and self.less_than_or_equal is None:
            errors.append("no less bound")
        return errors

    def __get_greater_bound(self) -> Optional[Number]:
        return self.greater_than if isinstance(self.greater_than, Number) else self.greater_than_or_equal

    def __get_less_bound(self) -> Optional[Number]:
        return self.less_than if isinstance(self.less_than, Number) else self.less_than_or_equal

    def get_between_range_error(self) -> Optional[str]:
        errors: list[str] = self.__get_common_range_errors()
        if not errors:
            greater_bound: Number = self.__get_greater_bound()
            less_bound: Number = self.__get_less_bound()
            if greater_bound >= less_bound:
                errors.append(f"greater bound ({greater_bound}) < less bound ({less_bound})")
        return " & ".join(errors) if errors else None

    def get_not_between_range_error(self) -> Optional[str]:
        errors: list[str] = self.__get_common_range_errors()
        if not errors:
            greater_bound: Number = self.__get_greater_bound()
            less_bound: Number = self.__get_less_bound()
            if greater_bound <= less_bound:
                errors.append(f"greater bound ({greater_bound}) > less bound ({less_bound})")
        return " & ".join(errors) if errors else None


class CheckYamlParser(ABC):
    @abstractmethod
    def get_check_type_names(self) -> list[str]:
        pass

    @abstractmethod
    def parse_check_yaml(
        self, check_type_name: str, check_yaml_object: YamlObject, column_yaml: Optional[ColumnYaml]
    ) -> Optional[CheckYaml]:
        pass


class CheckYaml(ABC):
    check_yaml_parsers: dict[str, CheckYamlParser] = {}

    @classmethod
    def register(cls, check_yaml_parser: CheckYamlParser) -> None:
        for check_type_name in check_yaml_parser.get_check_type_names():
            cls.check_yaml_parsers[check_type_name] = check_yaml_parser

    @classmethod
    def get_check_type_names(cls) -> list[str]:
        return list(cls.check_yaml_parsers.keys())

    @classmethod
    def parse_check_yaml(
        cls, check_type_name: str, check_body_yaml_object: YamlObject, column_yaml: Optional[ColumnYaml]
    ) -> Optional[CheckYaml]:
        if isinstance(check_type_name, str):
            check_yaml_parser: Optional[CheckYamlParser] = cls.check_yaml_parsers.get(check_type_name)
            if check_yaml_parser:
                return check_yaml_parser.parse_check_yaml(
                    check_type_name=check_type_name,
                    check_yaml_object=check_body_yaml_object,
                    column_yaml=column_yaml,
                )

    def __init__(self, type_name: str, check_yaml_object: YamlObject):
        self.check_yaml_object: YamlObject = check_yaml_object
        self.type_name: str = type_name
        self.name: Optional[str] = check_yaml_object.read_string_opt("name") if check_yaml_object else None
        qualifier = check_yaml_object.read_value("qualifier") if check_yaml_object else None
        self.qualifier: Optional[str] = str(qualifier) if qualifier is not None else None
        self.filter: Optional[str] = check_yaml_object.read_string_opt("filter") if check_yaml_object else None
        self.store_failed_rows: Optional[bool] = (
            check_yaml_object.read_bool_opt("store_failed_rows", default_value=False) if check_yaml_object else None
        )
        if self.filter:
            self.filter = self.filter.strip()
        self.attributes: dict[str, any] = check_yaml_object.read_object_opt("attributes", default_value={}).to_dict()


class ThresholdCheckYaml(CheckYaml):
    def __init__(self, type_name: str, check_yaml_object: YamlObject):
        super().__init__(type_name=type_name, check_yaml_object=check_yaml_object)
        self.metric: Optional[str] = None
        self.unit: Optional[str] = None
        self.threshold: Optional[ThresholdYaml] = None
        threshold_yaml_object: YamlObject = check_yaml_object.read_object_opt("threshold")
        if threshold_yaml_object:
            self.metric = self.read_metric(threshold_yaml_object)
            self.unit = self.read_unit(threshold_yaml_object)
            self.threshold = ThresholdYaml(threshold_yaml_object)

    def read_metric(self, check_yaml_object: YamlObject) -> Optional[str]:
        metric: Optional[str] = check_yaml_object.read_string_opt("metric")
        if metric:
            valid_metrics: list[str] = self.get_valid_metrics()
            if len(valid_metrics) == 0:
                logger.error(
                    msg=f"'metric' not allowed",
                    extra={ExtraKeys.LOCATION: check_yaml_object.create_location_from_yaml_dict_key("metric")},
                )
            elif metric not in valid_metrics:
                logger.error(
                    msg=f"'metric' must be in {valid_metrics}",
                    extra={ExtraKeys.LOCATION: check_yaml_object.create_location_from_yaml_dict_key("metric")},
                )
        return metric

    def get_valid_metrics(self) -> list[str]:
        return []

    def read_unit(self, check_yaml_object: YamlObject) -> Optional[str]:
        unit: Optional[str] = check_yaml_object.read_string_opt("unit")
        if unit:
            valid_units: list[str] = self.get_valid_units()
            if len(valid_units) == 0:
                logger.error(
                    msg=f"'unit' not allowed",
                    extra={ExtraKeys.LOCATION: check_yaml_object.create_location_from_yaml_dict_key("metric")},
                )
            if unit not in valid_units:
                logger.error(
                    msg=f"'metric' must be in {valid_units}, but was {unit}",
                    extra={ExtraKeys.LOCATION: check_yaml_object.create_location_from_yaml_dict_key("unit")},
                )
        return unit

    def get_valid_units(self) -> list[str]:
        return []


class ThresholdYaml:
    def __init__(self, threshold_yaml_object: YamlObject):
        self.must_be_greater_than: Optional[Number] = threshold_yaml_object.read_number_opt("must_be_greater_than")
        self.must_be_greater_than_or_equal: Optional[Number] = threshold_yaml_object.read_number_opt(
            "must_be_greater_than_or_equal"
        )
        self.must_be_less_than: Optional[Number] = threshold_yaml_object.read_number_opt("must_be_less_than")
        self.must_be_less_than_or_equal: Optional[Number] = threshold_yaml_object.read_number_opt(
            "must_be_less_than_or_equal"
        )
        self.must_be: Optional[Number] = threshold_yaml_object.read_number_opt("must_be")
        self.must_not_be: Optional[Number] = threshold_yaml_object.read_number_opt("must_not_be")
        self.must_be_between: Optional[RangeYaml] = RangeYaml.read_range_opt(threshold_yaml_object, "must_be_between")
        self.must_be_not_between: Optional[RangeYaml] = RangeYaml.read_range_opt(
            threshold_yaml_object, "must_be_not_between"
        )
        self.level: str = threshold_yaml_object.read_string_opt("level", default_value="fail")

    @classmethod
    def __config_count(cls, members: list[any]) -> int:
        return sum([0 if v is None else 1 for v in members])

    def has_any_configurations(self) -> bool:
        return (
            self.__config_count(
                [
                    self.must_be_greater_than,
                    self.must_be_greater_than_or_equal,
                    self.must_be_less_than,
                    self.must_be_less_than_or_equal,
                    self.must_be,
                    self.must_not_be,
                    self.must_be_between,
                    self.must_be_not_between,
                ]
            )
            > 0
        )

    def has_exactly_one_comparison(self) -> bool:
        comparator_count: int = self.__config_count(
            [
                self.must_be_greater_than,
                self.must_be_greater_than_or_equal,
                self.must_be_less_than,
                self.must_be_less_than_or_equal,
                self.must_be,
                self.must_not_be,
            ]
        )
        between_count: int = self.__config_count([self.must_be_between, self.must_be_not_between])
        return comparator_count == 1 and between_count == 0


class MissingAncValidityCheckYaml(ThresholdCheckYaml, MissingAndValidityYaml):
    def __init__(self, type_name: str, check_yaml_object: YamlObject):
        ThresholdCheckYaml.__init__(self, type_name=type_name, check_yaml_object=check_yaml_object)
        MissingAndValidityYaml.__init__(self, yaml_object=check_yaml_object)

    def get_valid_metrics(self) -> list[str]:
        return ["count", "percent"]



from __future__ import annotations

from abc import ABC
from datetime import timezone
from enum import Enum
from io import StringIO
from logging import LogRecord
from typing import Protocol

from ruamel.yaml import YAML
from soda_core.common.consistent_hash_builder import ConsistentHashBuilder
from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.env_config_helper import EnvConfigHelper
from soda_core.common.exceptions import (
    InvalidRegexException,
    SodaCoreException,
    get_exception_stacktrace,
)
from soda_core.common.logging_constants import Emoticons, ExtraKeys
from soda_core.common.logs import Location, Logs
from soda_core.common.metadata_types import SamplerType
from soda_core.common.soda_cloud import SodaCloud
from soda_core.common.soda_cloud_converter import map_sampler_type_from_dto
from soda_core.common.soda_cloud_dto import DatasetConfigurationDTO
from soda_core.common.sql_dialect import *
from soda_core.common.yaml import (
    ContractYamlSource,
    DataSourceYamlSource,
    SodaCloudYamlSource,
)
from soda_core.contracts.contract_verification import (
    Check,
    CheckOutcome,
    CheckResult,
    Contract,
    ContractVerificationResult,
    ContractVerificationSessionResult,
    ContractVerificationStatus,
    DataSource,
    Measurement,
    PostProcessingStage,
    PostProcessingStageState,
    SodaException,
    Threshold,
    YamlFileContentInfo,
)
from soda_core.contracts.impl.contract_yaml import (
    CheckYaml,
    ColumnYaml,
    ContractYaml,
    MissingAncValidityCheckYaml,
    MissingAndValidityYaml,
    RegexFormat,
    ThresholdYaml,
    ValidReferenceDataYaml,
)
from tabulate import tabulate

logger: logging.Logger = soda_logger


class ContractVerificationHandler(ABC):
    @abstractmethod
    def handle(
        self,
        contract_impl: ContractImpl,
        data_source_impl: Optional[DataSourceImpl],
        contract_verification_result: ContractVerificationResult,
        soda_cloud: SodaCloud,
        soda_cloud_send_results_response_json: dict,
        dwh_data_source_file_path: Optional[str] = None,
    ):
        pass

    @abstractmethod
    def provides_post_processing_stages(self) -> list[PostProcessingStage]:
        pass


class ContractVerificationHandlerRegistry(ABC):
    contract_verification_handlers: list[ContractVerificationHandler] = []
    post_processing_stages: dict[str, ContractVerificationHandler] = {}

    @classmethod
    def register(cls, verification_handler: ContractVerificationHandler) -> None:
        cls.contract_verification_handlers.append(verification_handler)
        for stage in verification_handler.provides_post_processing_stages():
            stage_name = stage.name
            if stage_name in cls.post_processing_stages:
                logger.warning(f"Overriding existing contract verification handler for check type {stage_name}")
            cls.post_processing_stages[stage_name] = verification_handler


class ContractVerificationSessionImpl:
    """Implements the contract verification session.

    @param contract_yaml_sources: The list of contract YAML sources to verify.
    @param only_validate_without_execute: If True, only validate the contracts without executing them.
    @param variables: The variables to use in the contract queries.
    @param data_timestamp: The timestamp of the data to use for the verification.
    @param data_source_impls: The data source implementations to use for the verification.
    @param data_source_yaml_sources: The data source YAML sources to use for the verification.
    @param soda_cloud_impl: The Soda Cloud implementation to use for the verification.
    @param soda_cloud_publish_results: If True, publish the results to Soda Cloud.
    @param soda_cloud_use_agent: If True, use the Soda Cloud agent for the verification.
    @param soda_cloud_verbose: If True, enable verbose logging for the Soda Cloud agent.
    @param soda_cloud_use_agent_blocking_timeout_in_minutes: The timeout for the Soda Cloud agent.
    @param dwh_data_source_file_path: The file path to the Diagnostics Warehouse data source.
    """

    @classmethod
    def execute(
        cls,
        contract_yaml_sources: list[ContractYamlSource],
        only_validate_without_execute: bool = False,
        variables: Optional[dict[str, str]] = None,
        data_timestamp: Optional[str] = None,
        data_source_impls: Optional[list[DataSourceImpl]] = None,
        data_source_yaml_sources: Optional[list[DataSourceYamlSource]] = None,
        soda_cloud_impl: Optional[SodaCloud] = None,
        soda_cloud_publish_results: bool = False,
        soda_cloud_use_agent: bool = False,
        soda_cloud_verbose: bool = False,
        soda_cloud_use_agent_blocking_timeout_in_minutes: int = 60,
        check_paths: Optional[list[str]] = None,
        dwh_data_source_file_path: Optional[str] = None,
    ):
        logs: Logs = Logs()

        # Validate input contract_yaml_sources
        assert isinstance(contract_yaml_sources, list)
        assert all(
            isinstance(contract_yaml_source, ContractYamlSource) for contract_yaml_source in contract_yaml_sources
        )

        # Validate input variables
        if variables is None:
            variables = {}
        else:
            assert isinstance(variables, dict)
            assert all(isinstance(k, str) and isinstance(v, (str, Number)) for k, v in variables.items())

        # Validate input data_source_impls
        if data_source_impls is None:
            data_source_impls = []
        else:
            assert isinstance(data_source_impls, list)
            assert all(isinstance(data_source_impl, DataSourceImpl) for data_source_impl in data_source_impls)

        # Validate input data_source_yaml_sources
        if data_source_yaml_sources is None:
            data_source_yaml_sources = []
        else:
            assert isinstance(data_source_yaml_sources, list)
            assert all(
                isinstance(data_source_yaml_source, DataSourceYamlSource) or soda_cloud_use_agent
                for data_source_yaml_source in data_source_yaml_sources
            )

        # Validate input soda_cloud_impl
        if soda_cloud_impl is not None:
            assert isinstance(soda_cloud_impl, SodaCloud)

        # Validate input soda_cloud_skip_publish
        assert isinstance(soda_cloud_publish_results, bool)

        # Validate input soda_cloud_use_agent
        assert isinstance(soda_cloud_use_agent, bool)

        # Validate input soda_cloud_use_agent_blocking_timeout_in_minutes
        assert isinstance(soda_cloud_use_agent_blocking_timeout_in_minutes, int)

        if check_paths is None:
            check_paths = []

        if soda_cloud_use_agent:
            contract_verification_results: list[ContractVerificationResult] = cls._execute_on_agent(
                contract_yaml_sources=contract_yaml_sources,
                variables=variables,
                soda_cloud_impl=soda_cloud_impl,
                soda_cloud_use_agent_blocking_timeout_in_minutes=soda_cloud_use_agent_blocking_timeout_in_minutes,
                soda_cloud_publish_results=soda_cloud_publish_results,
                soda_cloud_verbose=soda_cloud_verbose,
            )

        else:
            contract_verification_results: list[ContractVerificationResult] = cls._execute_locally(
                logs=logs,
                contract_yaml_sources=contract_yaml_sources,
                only_validate_without_execute=only_validate_without_execute,
                provided_variable_values=variables,
                data_timestamp=data_timestamp,
                data_source_impls=data_source_impls,
                data_source_yaml_sources=data_source_yaml_sources,
                soda_cloud_impl=soda_cloud_impl,
                soda_cloud_publish_results=soda_cloud_publish_results,
                check_paths=check_paths,
                dwh_data_source_file_path=dwh_data_source_file_path,
            )
        return ContractVerificationSessionResult(contract_verification_results=contract_verification_results)

    @classmethod
    def _execute_locally(
        cls,
        logs: Logs,
        contract_yaml_sources: list[ContractYamlSource],
        only_validate_without_execute: bool,
        provided_variable_values: dict[str, str],
        data_timestamp: Optional[str],
        data_source_impls: list[DataSourceImpl],
        data_source_yaml_sources: list[DataSourceYamlSource],
        soda_cloud_impl: Optional[SodaCloud],
        soda_cloud_publish_results: bool,
        check_paths: list[str],
        dwh_data_source_file_path: Optional[str] = None,
    ) -> list[ContractVerificationResult]:
        "Verifies a Contract locally."
        contract_verification_results: list[ContractVerificationResult] = []

        data_source_impls_by_name: dict[str, DataSourceImpl] = cls._build_data_source_impls_by_name(
            data_source_impls=data_source_impls,
            data_source_yaml_sources=data_source_yaml_sources,
            provided_variable_values=provided_variable_values,
        )

        opened_data_sources: list[DataSourceImpl] = []
        try:
            for contract_yaml_source in contract_yaml_sources:
                try:
                    contract_yaml: ContractYaml = ContractYaml.parse(
                        contract_yaml_source=contract_yaml_source,
                        provided_variable_values=provided_variable_values,
                        data_timestamp=data_timestamp,
                        primary_data_source_impl=data_source_impls_by_name.get("primary_datasource"),
                    )
                    data_source_name: str = (
                        contract_yaml.dataset[: contract_yaml.dataset.find("/")] if contract_yaml.dataset else None
                    )
                    data_source_impl: Optional[DataSourceImpl] = (
                        cls._get_data_source_impl(data_source_name, data_source_impls_by_name, opened_data_sources)
                        if (contract_yaml and data_source_name and not only_validate_without_execute)
                        else None
                    )
                    contract_impl: ContractImpl = ContractImpl(
                        contract_yaml=contract_yaml,
                        only_validate_without_execute=only_validate_without_execute,
                        data_timestamp=contract_yaml.data_timestamp,
                        execution_timestamp=contract_yaml.execution_timestamp,
                        data_source_impl=data_source_impl,
                        all_data_source_impls=data_source_impls_by_name,
                        soda_cloud=soda_cloud_impl,
                        publish_results=soda_cloud_publish_results,
                        logs=logs,
                        check_paths=check_paths,
                        dwh_data_source_file_path=dwh_data_source_file_path,
                    )
                    contract_verification_result: ContractVerificationResult = contract_impl.verify()
                    contract_verification_results.append(contract_verification_result)
                except Exception as e:
                    raise e
        finally:
            for data_source_impl in opened_data_sources:
                data_source_impl.close_connection()
        return contract_verification_results

    @classmethod
    def _build_data_source_impls_by_name(
        cls,
        data_source_impls: list[DataSourceImpl],
        data_source_yaml_sources: list[DataSourceYamlSource],
        provided_variable_values: dict[str, str],
    ) -> dict[str, DataSourceImpl]:
        data_source_impl_by_name: dict[str, DataSourceImpl] = (
            {data_source_impl.name: data_source_impl for data_source_impl in data_source_impls}
            if data_source_impls
            else {}
        )
        for data_source_yaml_source in data_source_yaml_sources:
            data_source_impl: DataSourceImpl = DataSourceImpl.from_yaml_source(
                data_source_yaml_source=data_source_yaml_source, provided_variable_values=provided_variable_values
            )
            data_source_impl_by_name[data_source_impl.name] = data_source_impl
        return data_source_impl_by_name

    @classmethod
    def _build_soda_cloud_impl(
        cls,
        soda_cloud_impl: Optional[SodaCloud],
        soda_cloud_yaml_source: Optional[SodaCloudYamlSource],
        provided_variable_values: dict[str, str],
    ) -> Optional[SodaCloud]:
        if soda_cloud_impl:
            return soda_cloud_impl
        if soda_cloud_yaml_source:
            return SodaCloud.from_yaml_source(
                soda_cloud_yaml_source=soda_cloud_yaml_source, provided_variable_values=provided_variable_values
            )
        return None

    @classmethod
    def _get_data_source_impl(
        cls,
        data_source_name: Optional[str],
        data_source_impl_by_name: dict[str, DataSourceImpl],
        opened_data_sources: list[DataSourceImpl],
    ) -> Optional[DataSourceImpl]:
        if data_source_name is None:
            return None
        data_source_impl: Optional[DataSourceImpl] = data_source_impl_by_name.get(data_source_name)
        if isinstance(data_source_impl, DataSourceImpl):
            if not data_source_impl.has_open_connection():
                data_source_impl.open_connection()
                opened_data_sources.append(data_source_impl)
            return data_source_impl
        else:
            logger.error(f"Data source '{data_source_name}' not found")

    @classmethod
    def _execute_on_agent(
        cls,
        contract_yaml_sources: list[ContractYamlSource],
        variables: dict[str, str],
        soda_cloud_impl: Optional[SodaCloud],
        soda_cloud_use_agent_blocking_timeout_in_minutes: int,
        soda_cloud_publish_results: bool,
        soda_cloud_verbose: bool,
    ) -> list[ContractVerificationResult]:
        "Verifies Contracts on the Soda Cloud agent."
        contract_verification_results: list[ContractVerificationResult] = []

        for contract_yaml_source in contract_yaml_sources:
            try:
                contract_yaml: ContractYaml = ContractYaml.parse(
                    contract_yaml_source=contract_yaml_source, provided_variable_values=variables
                )
                contract_verification_result: ContractVerificationResult = soda_cloud_impl.verify_contract_on_agent(
                    contract_yaml=contract_yaml,
                    variables=variables,
                    blocking_timeout_in_minutes=soda_cloud_use_agent_blocking_timeout_in_minutes,
                    publish_results=soda_cloud_publish_results,
                    verbose=soda_cloud_verbose,
                )
                contract_verification_results.append(contract_verification_result)
            except:
                logger.error(msg=f"Could not verify contract {contract_yaml_source}", exc_info=True)
        return contract_verification_results


class ContractImplExtension(Protocol):
    def __init__(self, contract_impl: ContractImpl):
        self.contract_impl: ContractImpl = contract_impl

    def parse_checks(self, contract_impl: ContractImpl) -> list[CheckImpl]:
        return []

    def build_queries(self, contract_impl: ContractImpl) -> list[Query]:
        return []


class ContractImpl:
    contract_impl_extensions: dict[str, type[ContractImplExtension]] = {}

    @classmethod
    def register_extension(cls, name: str, extension_cls: type[ContractImplExtension]) -> None:
        cls.contract_impl_extensions[name] = extension_cls

    def __init__(
        self,
        logs: Logs,
        contract_yaml: ContractYaml,
        only_validate_without_execute: bool,
        data_source_impl: Optional[DataSourceImpl],
        all_data_source_impls: dict[str, DataSourceImpl],
        data_timestamp: datetime,
        execution_timestamp: datetime,
        soda_cloud: Optional[SodaCloud],
        publish_results: bool,
        check_paths: list[str] = [],
        dwh_data_source_file_path: Optional[str] = None,
    ):
        self.logs: Logs = logs
        self.contract_yaml: ContractYaml = contract_yaml
        self.only_validate_without_execute: bool = only_validate_without_execute
        self.data_source_impl: DataSourceImpl = data_source_impl
        self.all_data_source_impls: dict[str, DataSourceImpl] = all_data_source_impls
        self.soda_cloud: Optional[SodaCloud] = soda_cloud
        self.publish_results: bool = publish_results
        self.soda_config = EnvConfigHelper()

        self.filter: Optional[str] = self.contract_yaml.filter
        self.check_paths: list[str] = check_paths

        self.started_timestamp: datetime = datetime.now(tz=timezone.utc)

        self.execution_timestamp: datetime = execution_timestamp
        self.data_timestamp: datetime = data_timestamp

        self.dataset_name: Optional[str] = None

        self.check_attributes: dict[str, any] = contract_yaml.check_attributes

        self.dataset_identifier = DatasetIdentifier.parse(contract_yaml.dataset)
        self.dataset_prefix: list[str] = self.dataset_identifier.prefixes
        self.dataset_name = self.dataset_identifier.dataset_name

        self.metrics_resolver: MetricsResolver = MetricsResolver()

        self.column_impls: list[ColumnImpl] = []
        self.check_impls: list[CheckImpl] = []

        # TODO replace usage of self.soda_qualified_dataset_name with self.dataset_identifier
        self.soda_qualified_dataset_name = contract_yaml.dataset
        # TODO replace usage of self.sql_qualified_dataset_name with self.dataset_identifier
        self.sql_qualified_dataset_name: Optional[str] = None

        self.datasource_warehouse: Optional[str] = None
        self.compute_warehouse: Optional[str] = None

        if data_source_impl:
            # TODO replace usage of self.sql_qualified_dataset_name with self.dataset_identifier
            self.sql_qualified_dataset_name = data_source_impl.sql_dialect.qualify_dataset_name(
                dataset_prefix=self.dataset_prefix, dataset_name=self.dataset_name
            )

            if data_source_impl.data_source_connection:
                if hasattr(data_source_impl.data_source_connection.connection_properties, "warehouse"):
                    self.datasource_warehouse = data_source_impl.data_source_connection.connection_properties.warehouse

                if self.datasource_warehouse is None:
                    self.datasource_warehouse = data_source_impl.get_current_warehouse()

        from soda_core.contracts.impl.check_types.row_count_check import (
            RowCountMetricImpl,
        )

        self.row_count_metric_impl: MetricImpl = self.metrics_resolver.resolve_metric(
            RowCountMetricImpl(contract_impl=self)
        )
        self.dataset_rows_tested: Optional[int] = None

        # Dataset defining CTE - used as basis for all queries in this contract
        self.cte = CTE("_soda_filtered_dataset").AS(
            [
                SELECT(STAR()),
                FROM(self.dataset_identifier.dataset_name, self.dataset_identifier.prefixes),
                WHERE.optional(SqlExpressionStr.optional(self.filter)),
            ]
        )
        # Optional sampler configuration. Is there a better place or way to store this?
        self.sampler_type: Optional[SamplerType] = None
        self.sampler_limit: Optional[Number] = None

        self.dataset_configuration: Optional[DatasetConfigurationDTO] = None
        if self.soda_cloud:
            self.dataset_configuration = self.soda_cloud.fetch_dataset_configuration(self.dataset_identifier)

        if self.dataset_configuration:
            if (
                self.dataset_configuration.test_row_sampler_configuration
                and self.dataset_configuration.test_row_sampler_configuration.enabled
                and self.dataset_configuration.test_row_sampler_configuration.test_row_sampler is not None
            ):
                self.sampler_type = map_sampler_type_from_dto(
                    self.dataset_configuration.test_row_sampler_configuration.test_row_sampler.type
                )
                self.sampler_limit = self.dataset_configuration.test_row_sampler_configuration.test_row_sampler.limit

            if self.dataset_configuration.compute_warehouse_override:
                self.compute_warehouse = self.dataset_configuration.compute_warehouse_override.name

        if self.should_apply_sampling:
            logger.info(
                f"Row sampling is enabled for dataset {self.dataset_identifier.to_string()} "
                f"with sampler config: type:'{self.dataset_configuration.test_row_sampler_configuration.test_row_sampler.type}', limit:'{self.dataset_configuration.test_row_sampler_configuration.test_row_sampler.limit}'"
            )

            # This modifies the CTE to include sampling by accessing the first element of the cte_query list, may be flaky. Consider adding a better way to modify queries, or change AST to a 3rd party library which may support it already.
            self.cte.cte_query[1] = self.cte.cte_query[1].SAMPLE(
                self.sampler_type,
                self.sampler_limit,
            )

        self.extensions: list[ContractImplExtension] = []
        for extension_cls in ContractImpl.contract_impl_extensions.values():
            try:
                extension = extension_cls(self)
                self.extensions.append(extension)
            except Exception as e:
                logger.error(
                    f"Error extending contract implementation with extension {extension_cls.__name__}: {e}",
                )

        self.column_impls: list[ColumnImpl] = self._parse_columns(contract_yaml=contract_yaml)
        self.check_impls: list[CheckImpl] = self._parse_checks(contract_yaml)

        dataset_check_impls: list[CheckImpl] = list(self.check_impls)
        column_check_impls: list[CheckImpl] = []
        for column_impl in self.column_impls:
            column_check_impls.extend(column_impl.check_impls)
        # For consistency and predictability, we want the checks eval and results in the same order as in the contract
        self.all_check_impls: list[CheckImpl] = (
            dataset_check_impls + column_check_impls
            if self._dataset_checks_came_before_columns_in_yaml()
            else column_check_impls + dataset_check_impls
        )

        self._verify_duplicate_identities(self.all_check_impls)
        self.metrics: list[MetricImpl] = self.metrics_resolver.get_resolved_metrics()

        self.queries: list[Query] = []
        if data_source_impl:
            self.queries = self._build_queries()

        self.dwh_data_source_file_path: Optional[str] = dwh_data_source_file_path

    @property
    def is_test_verification_on_agent(self) -> bool:
        return self.soda_config.is_running_on_agent and self.soda_config.is_contract_test_scan_definition_type

    @property
    def is_sampling_enabled(self) -> bool:
        return self.sampler_type is not None and self.sampler_limit is not None

    @property
    def should_apply_sampling(self) -> bool:
        return self.is_test_verification_on_agent and self.is_sampling_enabled

    def _dataset_checks_came_before_columns_in_yaml(self) -> Optional[bool]:
        contract_keys: list[str] = self.contract_yaml.contract_yaml_object.keys()
        if "checks" in contract_keys and "columns" in contract_keys:
            return contract_keys.index("checks") < contract_keys.index("columns")
        return None

    def _get_data_timestamp(
        self, resolved_variable_values: dict[str, str], soda_variable_values: dict[str, str], default: datetime
    ) -> Optional[datetime]:
        # Skipped for 'NOW' :)
        return None

    def _parse_checks(self, contract_yaml: ContractYaml) -> list[CheckImpl]:
        check_impls: list[CheckImpl] = []
        if contract_yaml.checks:
            for check_yaml in contract_yaml.checks:
                if check_yaml:
                    check = CheckImpl.parse_check(contract_impl=self, check_yaml=check_yaml)
                    check_impls.append(check)

        for extension in self.extensions:
            try:
                check_impls.extend(extension.parse_checks(contract_impl=self))
            except Exception as e:
                logger.error(f"Error parsing checks with extension {extension.__class__.__name__}: {e}")

        return check_impls

    def _build_queries(self) -> list[Query]:
        queries: list[Query] = []
        aggregation_metrics: list[AggregationMetricImpl] = []

        for check in self.all_check_impls:
            queries.extend(check.queries)

        for metric in self.metrics:
            # Only build aggregation queries for metrics of known origin. Extensions might build their own queries.
            # Known origin means that the metric does not have any specific datasource/dataset associated or the ones that are linked are the same as the contract's datasource and dataset.
            if isinstance(metric, AggregationMetricImpl):
                if (metric.data_source_impl is None and metric.dataset_identifier is None) or (
                    metric.data_source_impl == self.data_source_impl
                    and metric.dataset_identifier == self.dataset_identifier
                ):
                    aggregation_metrics.append(metric)

        from soda_core.contracts.impl.check_types.schema_check import SchemaQuery

        schema_queries: list[SchemaQuery] = []
        other_queries: list[SchemaQuery] = []
        for query in queries:
            if isinstance(query, SchemaQuery):
                schema_queries.append(query)
            else:
                other_queries.append(query)

        aggregation_queries: list[AggregationQuery] = []
        for aggregation_metric in aggregation_metrics:
            if len(aggregation_queries) == 0 or not aggregation_queries[-1].can_accept(aggregation_metric):
                aggregation_queries.append(
                    AggregationQuery(
                        cte=self.cte,
                        dataset_prefix=self.dataset_prefix,
                        dataset_name=self.dataset_name,
                        data_source_impl=self.data_source_impl,
                        logs=self.logs,
                    )
                )
            last_aggregation_query: AggregationQuery = aggregation_queries[-1]
            last_aggregation_query.append_aggregation_metric(aggregation_metric)

        all_queries: list[Query] = schema_queries + aggregation_queries + other_queries

        for extension in self.extensions:
            try:
                extension_queries: list[Query] = extension.build_queries(contract_impl=self)
                all_queries.extend(extension_queries)
            except Exception as e:
                logger.error(f"Error building queries with extension {extension.__class__.__name__}: {e}")

        return all_queries

    def _parse_columns(self, contract_yaml: ContractYaml) -> list[ColumnImpl]:
        columns: list[ColumnImpl] = []
        if contract_yaml.columns:
            for column_yaml in contract_yaml.columns:
                if column_yaml:
                    column = ColumnImpl(contract_impl=self, column_yaml=column_yaml)
                    columns.append(column)
        return columns

    def verify(self) -> ContractVerificationResult:
        if self.data_source_impl and self.soda_config.is_running_on_agent:
            self.data_source_impl.switch_warehouse(self.compute_warehouse, contract_impl=self)
        data_source: Optional[DataSource] = None
        check_results: list[CheckResult] = []
        measurements: list[Measurement] = []
        contract_verification_status: ContractVerificationStatus = ContractVerificationStatus.UNKNOWN
        dataset_rows_tested: Optional[int] = None

        verb: str = "Validating" if self.only_validate_without_execute else "Verifying"
        logger.info(
            f"{verb} contract {Emoticons.SCROLL} "
            f"{self.contract_yaml.contract_yaml_source.file_path} {Emoticons.FINGERS_CROSSED}"
        )

        if self.data_source_impl:
            data_source = self.data_source_impl.build_data_source()

        if self.logs.has_errors:
            contract_verification_status = ContractVerificationStatus.ERROR

        elif not self.only_validate_without_execute:
            # Executing the queries will set the value of the metrics linked to queries
            for query in self.queries:
                query_measurements: list[Measurement] = query.execute()
                measurements.extend(query_measurements)

            measurement_values: MeasurementValues = MeasurementValues(measurements)

            self.dataset_rows_tested = measurement_values.get_value(self.row_count_metric_impl)

            # Triggering the derived metrics to initialize their value based on their dependencies
            derived_metric_impls: list[DerivedMetricImpl] = [
                derived_metric for derived_metric in self.metrics if isinstance(derived_metric, DerivedMetricImpl)
            ]
            for derived_metric_impl in derived_metric_impls:
                measurement_values.derive_value(derived_metric_impl)

            if self.data_source_impl:
                # Evaluate the checks
                for check_impl in self.all_check_impls:
                    if check_impl.skip:
                        logger.info(f"Skipping evaluation of check at path '{check_impl.path}'")
                        check_result: CheckResult = CheckResult(
                            check=check_impl._build_check_info(), outcome=CheckOutcome.EXCLUDED
                        )
                    else:
                        check_result: CheckResult = check_impl.evaluate(measurement_values=measurement_values)
                    check_results.append(check_result)

            contract_verification_status = _get_contract_verification_status(self.logs.has_errors, check_results)

            log_lines = self.build_log_summary(
                soda_qualified_dataset_name=self.soda_qualified_dataset_name, check_results=check_results
            )
            for line in log_lines:
                logger.info(line)

        log_records: Optional[list[LogRecord]] = self.logs.pop_log_records()

        soda_cloud_file_id: Optional[str] = None
        sending_results_to_soda_cloud_failed: bool = False
        contract_yaml_source_str_original = self.contract_yaml.contract_yaml_source.yaml_str_original
        soda_cloud_response_json: Optional[dict] = None

        if self.soda_cloud and self.publish_results:
            soda_cloud_file_id = self.soda_cloud._upload_contract_yaml_file(contract_yaml_source_str_original)

        post_processing_stages: list[PostProcessingStage] = []
        for contract_verification_handler in ContractVerificationHandlerRegistry.post_processing_stages.values():
            post_processing_stages += contract_verification_handler.provides_post_processing_stages()

        contract_verification_result: ContractVerificationResult = ContractVerificationResult(
            contract=Contract(
                data_source_name=self.data_source_impl.name if self.data_source_impl else None,
                dataset_prefix=self.dataset_prefix,
                dataset_name=self.dataset_name,
                soda_qualified_dataset_name=self.soda_qualified_dataset_name,
                source=YamlFileContentInfo(
                    source_content_str=contract_yaml_source_str_original,
                    local_file_path=self.contract_yaml.contract_yaml_source.file_path,
                    soda_cloud_file_id=soda_cloud_file_id,
                ),
                dataset_id=None,  # Set this to None for now (default). Will be filled later when we get the dataset id from Soda Cloud
            ),
            data_source=data_source,
            data_timestamp=self.data_timestamp,
            started_timestamp=self.started_timestamp,
            ended_timestamp=datetime.now(tz=timezone.utc),
            measurements=measurements,
            check_results=check_results,
            sending_results_to_soda_cloud_failed=sending_results_to_soda_cloud_failed,
            status=contract_verification_status,
            log_records=log_records,
            post_processing_stages=post_processing_stages,
        )

        scan_id: Optional[str] = None
        if soda_cloud_file_id:
            # send_contract_result will use contract.source.soda_cloud_file_id
            soda_cloud_response_json = self.soda_cloud.send_contract_result(contract_verification_result)
            scan_id = soda_cloud_response_json.get("scanId") if soda_cloud_response_json else None
            if not scan_id:
                contract_verification_result.sending_results_to_soda_cloud_failed = True
            else:
                contract_verification_result.scan_id = scan_id
                # Put the dataset id in the contract object
                contract_verification_result.contract.dataset_id = self.__get_dataset_id(
                    soda_cloud_response_json, self.soda_qualified_dataset_name
                )
        else:
            logger.debug(f"Not sending results to Soda Cloud {Emoticons.CROSS_MARK}")

        for contract_verification_handler in ContractVerificationHandlerRegistry.contract_verification_handlers:
            try:
                contract_verification_handler.handle(
                    contract_impl=self,
                    data_source_impl=self.data_source_impl,
                    contract_verification_result=contract_verification_result,
                    soda_cloud=self.soda_cloud,
                    soda_cloud_send_results_response_json=soda_cloud_response_json,
                    dwh_data_source_file_path=self.dwh_data_source_file_path,
                )
            except Exception as e:
                logger.error(f"Error in contract verification handler: {e}", exc_info=True)
                self._handle_post_processing_failure(
                    scan_id=scan_id, exc=e, contract_verification_handler=contract_verification_handler
                )

        return contract_verification_result

    def __get_dataset_id(self, soda_cloud_response_json: dict, qualified_dataset_name: str) -> Optional[str]:
        # Find the dataset id for the given qualified dataset name
        for check in soda_cloud_response_json.get("checks", []):
            for datasets in check.get("datasets", []):
                dataset_dqn: Optional[str] = datasets.get("dqn")
                if dataset_dqn and dataset_dqn == qualified_dataset_name:
                    return datasets.get("id")
        return None

    def build_log_summary(self, soda_qualified_dataset_name: str, check_results: list[CheckResult]) -> list[str]:
        summary_lines: list[str] = []

        failed_count: int = 0
        warned_count: int = 0
        not_evaluated_count: int = 0
        passed_count: int = 0
        excluded_count: int = 0

        for check_result in check_results:
            if check_result.is_failed:
                failed_count += 1
            elif check_result.is_not_evaluated:
                not_evaluated_count += 1
            elif check_result.is_passed:
                passed_count += 1
            elif check_result.is_warned:
                warned_count += 1
            elif check_result.is_excluded:
                excluded_count += 1
        total_count: int = failed_count + not_evaluated_count + passed_count + warned_count + excluded_count

        error_count: int = len(self.logs.get_errors())

        table_lines = [
            ["Checks", total_count],
            ["Passed", passed_count, Emoticons.WHITE_CHECK_MARK],
        ]

        if failed_count > 0:
            table_lines.append(["Failed", failed_count, Emoticons.CROSS_MARK])
        else:
            table_lines.append(["Failed", failed_count, Emoticons.WHITE_CHECK_MARK])

        if warned_count > 0:
            table_lines.append(["Warned", warned_count, Emoticons.WARNING])
        else:
            table_lines.append(["Warned", warned_count, Emoticons.WHITE_CHECK_MARK])

        if not_evaluated_count > 0:
            table_lines.append(["Not Evaluated", not_evaluated_count, Emoticons.CROSS_MARK])
        else:
            table_lines.append(["Not Evaluated", not_evaluated_count, Emoticons.WHITE_CHECK_MARK])

        if excluded_count > 0:
            table_lines.append(["Excluded", excluded_count, Emoticons.QUESTION_MARK])
        else:
            table_lines.append(["Excluded", excluded_count, Emoticons.WHITE_CHECK_MARK])

        if error_count > 0:
            table_lines.append(["Runtime Errors", error_count, Emoticons.CROSS_MARK])
        else:
            table_lines.append(["Runtime Errors", error_count, Emoticons.WHITE_CHECK_MARK])

        summary_lines.append(f"\n### Contract results for {soda_qualified_dataset_name}")
        summary_lines.append(self.build_summary_table(check_results))

        overview_table = tabulate(table_lines, tablefmt="github", stralign="left")
        summary_lines.append(f"# Summary:\n{overview_table}\n")

        return [line for joined_line in summary_lines for line in joined_line.split("\n")]

    def build_summary_table(self, check_results: list[CheckResult]) -> str:
        overview_table_data = [check_result.log_table_row() for check_result in check_results]

        # Sort by column name, check name and check outcome
        overview_table_data.sort(key=lambda row: (row["Column"], row["Check"], row["Outcome"]))

        # Re-iterate rows data and remove column name if it is the same as the previous row
        previous_column_name: Optional[str] = None
        for row in overview_table_data:
            if previous_column_name == row["Column"]:
                row["Column"] = ""  # Clear column name if it is the same as the previous row
            else:
                previous_column_name = row["Column"]

        return tabulate(overview_table_data, headers="keys", tablefmt="grid")

    @classmethod
    def _verify_duplicate_identities(cls, all_check_impls: list[CheckImpl]):
        checks_by_identity: dict[str, CheckImpl] = {}
        for check_impl in all_check_impls:
            existing_check_impl: Optional[CheckImpl] = checks_by_identity.get(check_impl.identity)
            if existing_check_impl:
                original_location: Optional[Location] = existing_check_impl.check_yaml.check_yaml_object.location
                original_location_str: str = f" Original({original_location})" if original_location else ""
                duplicate_location: Optional[Location] = check_impl.check_yaml.check_yaml_object.location
                duplicate_location_str: str = f" Duplicate({duplicate_location})" if duplicate_location else ""
                logger.error(
                    msg=(
                        f"Duplicate identity {check_impl.build_identity_path()}."
                        f"{original_location_str}{duplicate_location_str}"
                    ),
                    extra={
                        ExtraKeys.LOCATION: duplicate_location,
                    },
                )
            checks_by_identity[check_impl.identity] = check_impl

    @classmethod
    def compute_data_quality_score(cls, total_failed_rows_count: int, total_rows_count: int) -> float:
        return 100 - (total_failed_rows_count * 100 / total_rows_count)

    def _handle_post_processing_failure(
        self, scan_id: Optional[str], exc: Exception, contract_verification_handler: ContractVerificationHandler
    ):
        if scan_id is None:
            logger.warning("Not sending post-processing stage updates to Soda Cloud - no scan ID")
            return
        if self.soda_cloud is None:
            logger.warning("Not sending post-processing stage updates to Soda Cloud - no Soda Cloud client")
            return
        for post_processing_stage in contract_verification_handler.provides_post_processing_stages():
            self.soda_cloud.post_processing_update(
                stage=post_processing_stage.name,
                scan_id=scan_id,
                state=PostProcessingStageState.FAILED,
                error=get_exception_stacktrace(exc),
            )


def _get_contract_verification_status(has_errors: bool, check_results: list[CheckResult]) -> ContractVerificationStatus:
    if has_errors:
        return ContractVerificationStatus.ERROR

    if any(check_result.outcome == CheckOutcome.FAILED for check_result in check_results):
        return ContractVerificationStatus.FAILED

    if any(check_result.outcome == CheckOutcome.WARN for check_result in check_results):
        return ContractVerificationStatus.WARNED

    if all(check_result.outcome == CheckOutcome.PASSED for check_result in check_results):
        return ContractVerificationStatus.PASSED

    return ContractVerificationStatus.UNKNOWN


class MeasurementValues:
    def __init__(self, measurements: list[Measurement]):
        self.metric_values_by_id: dict[str, any] = {
            measurement.metric_id: measurement.value for measurement in measurements
        }
        self.metric_ids_being_derived: set[str] = set()

    def get_value(self, metric_impl: MetricImpl) -> any:
        return self.metric_values_by_id.get(metric_impl.id)

    def derive_value(self, derived_metric_impl: DerivedMetricImpl) -> None:
        if derived_metric_impl.id not in self.metric_values_by_id:
            if derived_metric_impl.id in self.metric_ids_being_derived:
                logger.error("Bug: please report circular reference in derived metrics")
            else:
                self.metric_ids_being_derived.add(derived_metric_impl.id)
                for metric_dependency in derived_metric_impl.get_metric_dependencies():
                    if isinstance(metric_dependency, DerivedMetricImpl):
                        self.derive_value(metric_dependency)
                self.metric_ids_being_derived.remove(derived_metric_impl.id)
                value = derived_metric_impl.compute_derived_value(self)
                self.metric_values_by_id[derived_metric_impl.id] = value


class ColumnImpl:
    def __init__(self, contract_impl: ContractImpl, column_yaml: ColumnYaml):
        self.column_yaml = column_yaml
        self.missing_and_validity: MissingAndValidity = MissingAndValidity(missing_and_validity_yaml=column_yaml)
        self.check_impls: list[CheckImpl] = []
        if column_yaml.check_yamls:
            for check_yaml in column_yaml.check_yamls:
                if check_yaml:
                    check = CheckImpl.parse_check(
                        contract_impl=contract_impl,
                        column_impl=self,
                        check_yaml=check_yaml,
                    )
                    self.check_impls.append(check)


class ValidReferenceData:
    @classmethod
    def parse(cls, valid_reference_data_yaml: Optional[ValidReferenceDataYaml]) -> Optional[ValidReferenceData]:
        return (
            ValidReferenceData(valid_reference_data_yaml)
            if valid_reference_data_yaml and valid_reference_data_yaml.dataset and valid_reference_data_yaml.column
            else None
        )

    def __init__(self, valid_reference_data_yaml: ValidReferenceDataYaml):
        dataset_identifier: DatasetIdentifier = DatasetIdentifier.parse(valid_reference_data_yaml.dataset)
        self.data_source_name: str = dataset_identifier.data_source_name
        self.dataset_prefix: list[str] = dataset_identifier.prefixes
        self.dataset_name: str = dataset_identifier.dataset_name
        self.column: str = valid_reference_data_yaml.column


class MissingAndValidity:
    def __init__(self, missing_and_validity_yaml: MissingAndValidityYaml):
        self.missing_values: Optional[list] = missing_and_validity_yaml.missing_values
        self.missing_format: Optional[RegexFormat] = missing_and_validity_yaml.missing_format

        self.invalid_values: Optional[list] = missing_and_validity_yaml.invalid_values
        self.invalid_format: Optional[RegexFormat] = missing_and_validity_yaml.invalid_format
        self.valid_values: Optional[list] = missing_and_validity_yaml.valid_values
        self.valid_format: Optional[RegexFormat] = missing_and_validity_yaml.valid_format
        self.valid_min: Optional[Number] = missing_and_validity_yaml.valid_min
        self.valid_max: Optional[Number] = missing_and_validity_yaml.valid_max
        self.valid_length: Optional[int] = missing_and_validity_yaml.valid_length
        self.valid_min_length: Optional[int] = missing_and_validity_yaml.valid_min_length
        self.valid_max_length: Optional[int] = missing_and_validity_yaml.valid_max_length
        self.valid_reference_data: Optional[ValidReferenceData] = ValidReferenceData.parse(
            missing_and_validity_yaml.valid_reference_data
        )

    def is_missing_expr(self, column_name: str | COLUMN) -> SqlExpression:
        is_missing_clauses: list[SqlExpression] = [IS_NULL(column_name)]
        if isinstance(self.missing_values, list):
            literal_values = [LITERAL(value) for value in self.missing_values]
            is_missing_clauses.append(IN(column_name, literal_values))
        if isinstance(self.missing_format, RegexFormat) and isinstance(self.missing_format.regex, str):
            is_missing_clauses.append(REGEX_LIKE(column_name, self.missing_format.regex))
        return OR.optional(is_missing_clauses)

    def is_invalid_expr(self, column_name: str | COLUMN) -> Optional[SqlExpression]:
        invalid_clauses: list[SqlExpression] = []
        if isinstance(self.valid_values, list):
            literal_values = [LITERAL(value) for value in self.valid_values if value is not None]
            if None in self.valid_values:
                invalid_clauses.append(AND([NOT(IN(column_name, literal_values)), IS_NOT_NULL(column_name)]))
            elif not self.valid_values:
                invalid_clauses.append(AND([LITERAL(True)]))
            else:
                invalid_clauses.append(NOT(IN(column_name, literal_values)))
        if isinstance(self.invalid_values, list):
            literal_values = [LITERAL(value) for value in self.invalid_values if value is not None]
            if None in self.invalid_values:
                invalid_clauses.append(AND([IN(column_name, literal_values), IS_NULL(column_name)]))
            elif not self.invalid_values:
                invalid_clauses.append(AND([LITERAL(False)]))
            else:
                invalid_clauses.append(IN(column_name, literal_values))
        if isinstance(self.valid_format, RegexFormat) and isinstance(self.valid_format.regex, str):
            invalid_clauses.append(NOT(REGEX_LIKE(column_name, self.valid_format.regex)))
        if isinstance(self.valid_min, Number) or isinstance(self.valid_min, str):
            invalid_clauses.append(LT(column_name, LITERAL(self.valid_min)))
        if isinstance(self.valid_max, Number) or isinstance(self.valid_max, str):
            invalid_clauses.append(GT(column_name, LITERAL(self.valid_max)))
        if isinstance(self.invalid_format, RegexFormat) and isinstance(self.invalid_format.regex, str):
            invalid_clauses.append(REGEX_LIKE(column_name, self.invalid_format.regex))
        if isinstance(self.valid_length, int):
            invalid_clauses.append(NEQ(LENGTH(column_name), LITERAL(self.valid_length)))
        if isinstance(self.valid_min_length, int):
            invalid_clauses.append(LT(LENGTH(column_name), LITERAL(self.valid_min_length)))
        if isinstance(self.valid_max_length, int):
            invalid_clauses.append(GT(LENGTH(column_name), LITERAL(self.valid_max_length)))
        return OR.optional(invalid_clauses)

    def is_valid_expr(self, column_name: str | COLUMN) -> SqlExpression:
        return NOT.optional(OR.optional([self.is_missing_expr(column_name), self.is_invalid_expr(column_name)]))

    @classmethod
    def __apply_default(cls, self_value, default_value) -> any:
        if self_value is not None:
            return self_value
        return default_value

    def apply_column_defaults(self, column_impl: ColumnImpl) -> None:
        if not column_impl:
            return
        column_defaults: MissingAndValidity = column_impl.missing_and_validity
        if not column_defaults:
            return

        check_has_missing: bool = self.has_missing_configurations()
        self.missing_values = self.missing_values if check_has_missing else column_defaults.missing_values
        self.missing_format = self.missing_format if check_has_missing else column_defaults.missing_format

        check_has_validity: bool = self.has_validity_configurations()
        self.invalid_values = self.invalid_values if check_has_validity else column_defaults.invalid_values
        self.invalid_format = self.invalid_format if check_has_validity else column_defaults.invalid_format
        self.valid_values = self.valid_values if check_has_validity else column_defaults.valid_values
        self.valid_format = self.valid_format if check_has_validity else column_defaults.valid_format
        self.valid_min = self.valid_min if check_has_validity else column_defaults.valid_min
        self.valid_max = self.valid_max if check_has_validity else column_defaults.valid_max
        self.valid_length = self.valid_length if check_has_validity else column_defaults.valid_length
        self.valid_min_length = self.valid_min_length if check_has_validity else column_defaults.valid_min_length
        self.valid_max_length = self.valid_max_length if check_has_validity else column_defaults.valid_max_length
        self.valid_reference_data = (
            self.valid_reference_data if check_has_validity else column_defaults.valid_reference_data
        )

    def has_missing_configurations(self) -> bool:
        return self.missing_values is not None or self.missing_format is not None

    def has_validity_configurations(self) -> bool:
        return any(
            cfg is not None
            for cfg in [
                self.invalid_values,
                self.invalid_format,
                self.valid_values,
                self.valid_format,
                self.valid_min,
                self.valid_max,
                self.valid_length,
                self.valid_min_length,
                self.valid_max_length,
                self.valid_reference_data,
            ]
        )

    def has_reference_data(self) -> bool:
        return isinstance(self.valid_reference_data, ValidReferenceData)


class MetricsResolver:
    def __init__(self):
        self.metrics: list[MetricImpl] = []

    def resolve_metric(self, metric_impl: MetricImpl) -> MetricImpl:
        existing_metric_impl: Optional[MetricImpl] = next((m for m in self.metrics if m == metric_impl), None)
        if existing_metric_impl:
            return existing_metric_impl
        else:
            self.metrics.append(metric_impl)
            return metric_impl

    def get_resolved_metrics(self) -> list[MetricImpl]:
        return self.metrics


class ThresholdType(Enum):
    SINGLE_COMPARATOR = "single_comparator"
    INNER_RANGE = "inner_range"
    OUTER_RANGE = "outer_range"


class ThresholdLevel(Enum):
    FAIL = "fail"
    WARN = "warn"

    @classmethod
    def from_str(cls, level_str: str) -> ThresholdLevel:
        level_str_lower = level_str.lower()
        if level_str_lower == "fail":
            return ThresholdLevel.FAIL
        elif level_str_lower == "warn":
            return ThresholdLevel.WARN
        else:
            logger.warning(f"Unknown threshold level '{level_str}', defaulting to 'fail'")
            return ThresholdLevel.FAIL


class ThresholdImpl:
    @classmethod
    def create(
        cls, threshold_yaml: ThresholdYaml, default_threshold: Optional[ThresholdImpl] = None
    ) -> Optional[ThresholdImpl]:
        if threshold_yaml is None:
            if default_threshold:
                return default_threshold
            else:
                logger.error(f"Threshold required, but not specified")
                return None

        threshold_level: ThresholdLevel = ThresholdLevel.from_str(threshold_yaml.level)

        if default_threshold:
            default_threshold.level = threshold_level

        if not threshold_yaml.has_any_configurations():
            if default_threshold:
                return default_threshold
            logger.error(f"Threshold required, but not specified")
            return None

        if threshold_yaml.has_exactly_one_comparison():
            return ThresholdImpl(
                type=ThresholdType.SINGLE_COMPARATOR,
                level=threshold_level,
                must_be_greater_than=threshold_yaml.must_be_greater_than,
                must_be_greater_than_or_equal=threshold_yaml.must_be_greater_than_or_equal,
                must_be_less_than=threshold_yaml.must_be_less_than,
                must_be_less_than_or_equal=threshold_yaml.must_be_less_than_or_equal,
                must_be=threshold_yaml.must_be,
                must_not_be=threshold_yaml.must_not_be,
            )

        elif threshold_yaml.must_be_between:
            range_error: Optional[str] = threshold_yaml.must_be_between.get_between_range_error()
            if range_error:
                logger.error(f"Invalid between threshold range: {range_error}")
                return None
            else:
                return ThresholdImpl(
                    type=ThresholdType.INNER_RANGE,
                    level=threshold_level,
                    must_be_greater_than=threshold_yaml.must_be_between.greater_than,
                    must_be_greater_than_or_equal=threshold_yaml.must_be_between.greater_than_or_equal,
                    must_be_less_than=threshold_yaml.must_be_between.less_than,
                    must_be_less_than_or_equal=threshold_yaml.must_be_between.less_than_or_equal,
                )

        elif threshold_yaml.must_be_not_between:
            range_error: Optional[str] = threshold_yaml.must_be_not_between.get_not_between_range_error()
            if range_error:
                logger.error(f"Invalid not between threshold range: {range_error}")
                return None
            else:
                return ThresholdImpl(
                    type=ThresholdType.OUTER_RANGE,
                    level=threshold_level,
                    must_be_greater_than=threshold_yaml.must_be_not_between.greater_than,
                    must_be_greater_than_or_equal=threshold_yaml.must_be_not_between.greater_than_or_equal,
                    must_be_less_than=threshold_yaml.must_be_not_between.less_than,
                    must_be_less_than_or_equal=threshold_yaml.must_be_not_between.less_than_or_equal,
                )

    def __init__(
        self,
        type: ThresholdType,
        level: ThresholdLevel = ThresholdLevel.FAIL,
        must_be_greater_than: Optional[Number] = None,
        must_be_greater_than_or_equal: Optional[Number] = None,
        must_be_less_than: Optional[Number] = None,
        must_be_less_than_or_equal: Optional[Number] = None,
        must_be: Optional[Number] = None,
        must_not_be: Optional[Number] = None,
    ):
        self.type: ThresholdType = type
        self.level: ThresholdLevel = level
        self.must_be_greater_than: Optional[Number] = must_be_greater_than
        self.must_be_greater_than_or_equal: Optional[Number] = must_be_greater_than_or_equal
        self.must_be_less_than: Optional[Number] = must_be_less_than
        self.must_be_less_than_or_equal: Optional[Number] = must_be_less_than_or_equal
        self.must_be: Optional[Number] = must_be
        self.must_not_be: Optional[Number] = must_not_be

    def to_threshold_info(self) -> Threshold:
        if self.must_be is None and self.must_not_be is None:
            return Threshold(
                level=self.level.value,
                must_be_greater_than=self.must_be_greater_than,
                must_be_greater_than_or_equal=self.must_be_greater_than_or_equal,
                must_be_less_than=self.must_be_less_than,
                must_be_less_than_or_equal=self.must_be_less_than_or_equal,
            )
        elif self.must_be is not None:
            return Threshold(
                level=self.level.value,
                must_be=self.must_be,
            )
        elif self.must_not_be is not None:
            return Threshold(
                level=self.level.value,
                must_not_be=self.must_not_be,
            )

    @classmethod
    def get_metric_name(cls, metric_name: str, column_impl: Optional[ColumnImpl]) -> str:
        if column_impl:
            return f"{metric_name}({column_impl.column_yaml.name})"
        else:
            return metric_name

    def get_assertion_summary(self, metric_name: str) -> str:
        """
        For ease of reading, thresholds always list small values lef and big values right (where applicable).
        Eg '0 < metric_name', 'metric_name < 25', '0 <= metric_name < 25'
        """
        if self.type == ThresholdType.SINGLE_COMPARATOR:
            if isinstance(self.must_be_greater_than, Number):
                return f"{self.must_be_greater_than} < {metric_name}"
            if isinstance(self.must_be_greater_than_or_equal, Number):
                return f"{self.must_be_greater_than_or_equal} <= {metric_name}"
            if isinstance(self.must_be_less_than, Number):
                return f"{metric_name} < {self.must_be_less_than}"
            if isinstance(self.must_be_less_than_or_equal, Number):
                return f"{metric_name} <= {self.must_be_less_than_or_equal}"
            if isinstance(self.must_be, Number):
                return f"{metric_name} = {self.must_be}"
            if isinstance(self.must_not_be, Number):
                return f"{metric_name} != {self.must_not_be}"
        elif self.type == ThresholdType.INNER_RANGE or self.type == ThresholdType.OUTER_RANGE:
            gt_comparator: str = " < " if isinstance(self.must_be_greater_than, Number) else " <= "
            gt_bound: str = (
                str(self.must_be_greater_than)
                if isinstance(self.must_be_greater_than, Number)
                else str(self.must_be_greater_than_or_equal)
            )
            lt_comparator: str = " < " if isinstance(self.must_be_less_than, Number) else " <= "
            lt_bound: str = (
                str(self.must_be_less_than)
                if isinstance(self.must_be_less_than, Number)
                else str(self.must_be_less_than_or_equal)
            )
            if self.type == ThresholdType.INNER_RANGE:
                return f"{gt_bound}{gt_comparator}{metric_name}{lt_comparator}{lt_bound}"
            else:
                return f"{metric_name}{lt_comparator}{lt_bound} or {gt_bound}{gt_comparator}{metric_name}"

    def passes(self, value: Number) -> bool:
        is_greater_than_ok: bool = (self.must_be_greater_than is None or value > self.must_be_greater_than) and (
            self.must_be_greater_than_or_equal is None or value >= self.must_be_greater_than_or_equal
        )
        is_less_than_ok: bool = (self.must_be_less_than is None or value < self.must_be_less_than) and (
            self.must_be_less_than_or_equal is None or value <= self.must_be_less_than_or_equal
        )
        if self.type == ThresholdType.OUTER_RANGE:
            return is_greater_than_ok or is_less_than_ok
        else:
            return (
                is_greater_than_ok
                and is_less_than_ok
                and (not (isinstance(self.must_be, Number) or isinstance(self.must_be, str)) or value == self.must_be)
                and (
                    not (isinstance(self.must_not_be, Number) or isinstance(self.must_not_be, str))
                    or value != self.must_not_be
                )
            )


class CheckParser(ABC):
    @abstractmethod
    def get_check_type_names(self) -> list[str]:
        pass

    @abstractmethod
    def parse_check(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: CheckYaml,
    ) -> Optional[CheckImpl]:
        pass


class CheckImpl:
    check_parsers: dict[str, CheckParser] = {}

    @classmethod
    def register(cls, check_parser: CheckParser) -> None:
        for check_type_name in check_parser.get_check_type_names():
            cls.check_parsers[check_type_name] = check_parser

    @classmethod
    def get_check_type_names(cls) -> list[str]:
        return list(cls.check_parsers.keys())

    @classmethod
    def parse_check(
        cls,
        contract_impl: ContractImpl,
        check_yaml: CheckYaml,
        column_impl: Optional[ColumnImpl] = None,
    ) -> Optional[CheckImpl]:
        if isinstance(check_yaml.type_name, str):
            check_parser: Optional[CheckParser] = cls.check_parsers.get(check_yaml.type_name)
            if check_parser:
                check_impl = check_parser.parse_check(
                    contract_impl=contract_impl,
                    column_impl=column_impl,
                    check_yaml=check_yaml,
                )

                if not check_impl.skip:
                    check_impl.setup_metrics(
                        contract_impl=contract_impl,
                        column_impl=column_impl,
                        check_yaml=check_yaml,
                    )

                return check_impl
            else:
                logger.error(f"Unknown check type '{check_yaml.type_name}'")

    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: CheckYaml,
    ):
        self.logs: Logs = contract_impl.logs

        self.contract_impl: ContractImpl = contract_impl
        self.check_yaml: CheckYaml = check_yaml
        self.name: str = self._get_name_with_default(check_yaml)
        self.column_impl: Optional[ColumnImpl] = column_impl
        self.type: str = check_yaml.type_name
        self.identity: str = self._build_identity(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_type=check_yaml.type_name,
            qualifier=check_yaml.qualifier,
        )

        self.threshold: Optional[ThresholdImpl] = None
        self.metrics: list[MetricImpl] = []
        self.queries: list[Query] = []
        self.skip: bool = False

        if contract_impl.check_paths:
            if self.path not in contract_impl.check_paths:
                self.skip = True

        # Merge check attributes with contract attributes
        self.attributes: dict[str, any] = {**contract_impl.check_attributes, **check_yaml.attributes}

    __DEFAULT_CHECK_NAMES_BY_TYPE: dict[str, str] = {
        "schema": "Schema matches expected structure",
        "row_count": "Row count meets expected threshold",
        "freshness": "Data is fresh",
        "missing": "No missing values",
        "invalid": "No invalid values",
        "duplicate": "No duplicate values",
        "aggregate": "Metric function meets threshold",
        "metric": "Metric meets threshold",
        "failed_rows": "No rows violating the condition",
    }

    @property
    def path(self) -> str:
        parts: list[str] = []

        column_name = self.column_impl.column_yaml.name if self.column_impl else None
        if column_name:
            parts.append("columns")
            parts.append(column_name)

        parts.append("checks")
        parts.append(self.type)

        qualifier = self.check_yaml.qualifier
        if qualifier:
            parts.append(qualifier)

        return ".".join(parts)

    def _get_name_with_default(self, check_yaml: CheckYaml) -> str:
        if isinstance(check_yaml.name, str):
            return check_yaml.name
        default_check_name: Optional[str] = self.__DEFAULT_CHECK_NAMES_BY_TYPE.get(check_yaml.type_name)
        if isinstance(default_check_name, str):
            return default_check_name
        return check_yaml.type_name

    def _resolve_metric(self, metric_impl: MetricImpl) -> MetricImpl:
        resolved_metric_impl: MetricImpl = self.contract_impl.metrics_resolver.resolve_metric(metric_impl)
        self.metrics.append(resolved_metric_impl)
        return resolved_metric_impl

    @abstractmethod
    def setup_metrics(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: CheckYaml,
    ) -> None:
        pass

    @abstractmethod
    def evaluate(self, measurement_values: MeasurementValues) -> CheckResult:
        pass

    def _build_check_info(self) -> Check:
        return Check(
            type=self.type,
            qualifier=self.check_yaml.qualifier,
            name=self.name,
            path=self.path,
            identity=self.identity,
            definition=self._build_definition(),
            column_name=self.column_impl.column_yaml.name if self.column_impl else None,
            contract_file_line=self.check_yaml.check_yaml_object.location.line,
            contract_file_column=self.check_yaml.check_yaml_object.location.column,
            threshold=self._build_threshold(),
            attributes=self.attributes,
            location=self.check_yaml.check_yaml_object.location,
        )

    @classmethod
    def _build_identity(
        cls, contract_impl: ContractImpl, column_impl: Optional[ColumnImpl], check_type: str, qualifier: Optional[str]
    ) -> str:
        identity_hash_builder: ConsistentHashBuilder = ConsistentHashBuilder(8)
        if contract_impl.data_source_impl:
            identity_hash_builder.add_property("dso", contract_impl.data_source_impl.name)
        identity_hash_builder.add_property("pr", contract_impl.dataset_prefix)
        identity_hash_builder.add_property("ds", contract_impl.dataset_name)
        identity_hash_builder.add_property("c", column_impl.column_yaml.name if column_impl else None)
        identity_hash_builder.add_property("t", check_type)
        identity_hash_builder.add_property("q", qualifier)

        return identity_hash_builder.get_hash()

    def build_identity_path(self) -> str:
        parts: list[Optional[str]] = [
            self.contract_impl.contract_yaml.contract_yaml_source.file_path,
            self.column_impl.column_yaml.name if self.column_impl else None,
            self.type,
            self.check_yaml.qualifier if self.check_yaml else None,
        ]
        parts = [p for p in parts if p is not None]
        return "/".join(parts)

    def _build_definition(self) -> str:
        contract_dict: dict = {}
        if self.contract_impl.contract_yaml.filter:
            contract_dict["filter"] = self.contract_impl.contract_yaml.filter

        check_dict: dict = self.check_yaml.check_yaml_object.yaml_dict

        if self.column_impl:
            contract_dict["columns"] = [{"name": self.column_impl.column_yaml.name, "checks": [check_dict]}]
        else:
            contract_dict["checks"] = [check_dict]

        text_stream = StringIO()
        yaml = YAML()
        yaml.dump(contract_dict, text_stream)
        text_stream.seek(0)
        return text_stream.read()

    def _build_threshold(self) -> Optional[Threshold]:
        return self.threshold.to_threshold_info() if self.threshold else None

    def get_threshold_metric_impl(self) -> Optional[MetricImpl]:
        """
        Used in extensions
        """
        raise SodaException(f"Check type '{self.type}' does not support get_threshold_metric_impl'")

    def evaluate_threshold(self, value) -> CheckOutcome:
        outcome: CheckOutcome = CheckOutcome.NOT_EVALUATED

        if self.threshold and isinstance(value, Number):
            if self.threshold.passes(value):
                outcome = CheckOutcome.PASSED
            else:
                if self.threshold.level == ThresholdLevel.WARN:
                    outcome = CheckOutcome.WARN
                else:
                    outcome = CheckOutcome.FAILED

        return outcome


class MissingAndValidityCheckImpl(CheckImpl):
    def __init__(
        self, contract_impl: ContractImpl, column_impl: Optional[ColumnImpl], check_yaml: MissingAncValidityCheckYaml
    ):
        super().__init__(contract_impl, column_impl, check_yaml)
        self.missing_and_validity: MissingAndValidity = MissingAndValidity(missing_and_validity_yaml=check_yaml)
        self.missing_and_validity.apply_column_defaults(column_impl)


class MetricImpl:
    def __init__(
        self,
        contract_impl: ContractImpl,
        metric_type: str,
        column_impl: Optional[ColumnImpl] = None,
        check_filter: Optional[str] = None,
        missing_and_validity: Optional[MissingAndValidity] = None,
        # Associate metric with a non-contract data source if needed. Build queries accordingly.
        data_source_impl: Optional[DataSourceImpl] = None,
        # Associate metric with a non-contract dataset if needed. Build queries accordingly.
        dataset_identifier: Optional[DatasetIdentifier] = None,
    ):
        self.contract_impl: ContractImpl = contract_impl
        self.column_impl: Optional[ColumnImpl] = column_impl
        self.type: str = metric_type
        self.check_filter: Optional[str] = check_filter
        self.missing_and_validity: Optional[MissingAndValidity] = missing_and_validity
        self.dataset_identifier = dataset_identifier or contract_impl.dataset_identifier

        self.data_source_impl: Optional[DataSourceImpl] = None
        if self.contract_impl.data_source_impl:
            self.data_source_impl = self.contract_impl.data_source_impl
        if data_source_impl:
            self.data_source_impl = data_source_impl

        self.id: str = self._build_id()

    def _build_id(self) -> str:
        hash_builder: ConsistentHashBuilder = ConsistentHashBuilder(hash_string_length=8)
        id_properties: dict[str, any] = self._get_id_properties()
        for k, v in id_properties.items():
            hash_builder.add_property(k, v)
        return hash_builder.get_hash()

    def _get_id_properties(self) -> dict[str, any]:
        id_properties: dict[str, any] = {"type": self.type}

        if self.data_source_impl:
            id_properties["data_source"] = self.data_source_impl.name
        id_properties["dataset"] = self.dataset_identifier.dataset_name
        if self.column_impl:
            id_properties["column"] = self.column_impl.column_yaml.name
        if self.check_filter:
            id_properties["check_filter"] = self.check_filter
        if self.missing_and_validity:
            id_properties["missing_values"] = self.missing_and_validity.missing_values
            if self.missing_and_validity.missing_format:
                id_properties["missing_format_regex"] = self.missing_and_validity.missing_format.regex
            id_properties["invalid_values"] = self.missing_and_validity.invalid_values
            if self.missing_and_validity.invalid_format:
                id_properties["invalid_format_regex"] = self.missing_and_validity.invalid_format.regex
            id_properties["valid_values"] = self.missing_and_validity.valid_values
            if self.missing_and_validity.valid_format:
                id_properties["valid_format"] = self.missing_and_validity.valid_format.regex
            id_properties["valid_min"] = self.missing_and_validity.valid_min
            id_properties["valid_max"] = self.missing_and_validity.valid_max
            id_properties["valid_length"] = self.missing_and_validity.valid_length
            id_properties["valid_min_length"] = self.missing_and_validity.valid_min_length
            id_properties["valid_max_length"] = self.missing_and_validity.valid_max_length
            if self.missing_and_validity.valid_reference_data:
                id_properties["ref_prefix"] = self.missing_and_validity.valid_reference_data.dataset_prefix
                id_properties["ref_name"] = self.missing_and_validity.valid_reference_data.dataset_name
                id_properties["ref_column"] = self.missing_and_validity.valid_reference_data.column

        return id_properties

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return self.id == other.id

    @abstractmethod
    def sql_condition_expression(self) -> Optional[SqlExpression]:
        pass


class AggregationMetricImpl(MetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        metric_type: str,
        column_impl: Optional[ColumnImpl] = None,
        check_filter: Optional[str] = None,
        missing_and_validity: Optional[MissingAndValidity] = None,
        data_source_impl: Optional[DataSourceImpl] = None,
        dataset_identifier: Optional[DatasetIdentifier] = None,
    ):
        super().__init__(
            contract_impl=contract_impl,
            metric_type=metric_type,
            column_impl=column_impl,
            check_filter=check_filter,
            missing_and_validity=missing_and_validity,
            data_source_impl=data_source_impl,
            dataset_identifier=dataset_identifier,
        )

    @abstractmethod
    def sql_expression(self) -> SqlExpression:
        pass

    @abstractmethod
    def sql_condition_expression(self) -> SqlExpression:
        """
        Used in extensions
        """

    def convert_db_value(self, value: any) -> any:
        return value

    def get_short_description(self) -> str:
        return self.type


class DerivedMetricImpl(MetricImpl, ABC):
    @abstractmethod
    def get_metric_dependencies(self) -> list[MetricImpl]:
        pass

    @abstractmethod
    def compute_derived_value(self, measurement_values: MeasurementValues) -> Number:
        pass

    def _get_id_properties(self) -> dict[str, any]:
        id_properties: dict[str, any] = super()._get_id_properties()
        for index, metric_dependency in enumerate(self.get_metric_dependencies()):
            id_properties[str(index)] = metric_dependency.id
        return id_properties


class DerivedPercentageMetricImpl(DerivedMetricImpl):
    def __init__(self, metric_type: str, fraction_metric_impl: MetricImpl, total_metric_impl: MetricImpl):
        self.fraction_metric_impl: MetricImpl = fraction_metric_impl
        self.total_metric_impl: MetricImpl = total_metric_impl
        # Mind the ordering as the self._build_id() must come last
        super().__init__(
            contract_impl=fraction_metric_impl.contract_impl,
            column_impl=fraction_metric_impl.column_impl,
            metric_type=metric_type,
            check_filter=None,
        )

    def get_metric_dependencies(self) -> list[MetricImpl]:
        return [self.fraction_metric_impl, self.total_metric_impl]

    def compute_derived_value(self, measurement_values: MeasurementValues) -> Number:
        fraction: int = measurement_values.get_value(self.fraction_metric_impl)
        if fraction is None:
            return 0
        total: int = measurement_values.get_value(self.total_metric_impl)
        return (fraction * 100 / total) if total != 0 else 0


class ValidCountMetric(AggregationMetricImpl):
    """TODO -- 3/10/2025: this metric is not used anywhere in the codebase, it's not clear if it's needed."""

    def __init__(self, contract_impl: ContractImpl, column_impl: ColumnImpl, check_impl: MissingAndValidityCheckImpl):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            metric_type="valid_count",
            check_filter=check_impl.check_yaml.filter,
            missing_and_validity=check_impl.missing_and_validity,
        )

    def sql_expression(self) -> SqlExpression:
        column_name = self.column_impl.column_yaml.name
        filters: list = [SqlExpressionStr.optional(self.check_filter)]
        if self.missing_and_validity:
            filters.append(NOT(self.missing_and_validity.is_missing_expr(column_name)))
            filters.append(NOT.optional(self.missing_and_validity.is_invalid_expr(column_name)))
        if filters:
            return SUM(CASE_WHEN(AND.optional(filters), LITERAL(1)))
        else:
            return COUNT(column_name)

    def convert_db_value(self, value) -> int:
        # Note: expression SUM(CASE WHEN "id" IS NULL THEN 1 ELSE 0 END) gives NULL / None as a result if
        # there are no rows
        return int(value) if value is not None else 0


class Query(ABC):
    def __init__(
        self, data_source_impl: Optional[DataSourceImpl], metrics: list[MetricImpl], sql: Optional[str] = None
    ):
        self.data_source_impl: DataSourceImpl = data_source_impl
        self.metrics: list[MetricImpl] = metrics
        self.sql: Optional[str] = sql

    def build_sql(self) -> str:
        """
        Signals the query building process is done.  The framework assumes that after this call the self.sql is initialized with the query string.
        """
        return self.sql

    @abstractmethod
    def execute(self) -> list[Measurement]:
        pass


class AggregationQuery(Query):
    def __init__(
        self,
        cte: CTE,
        dataset_prefix: list[str],
        dataset_name: str,
        data_source_impl: Optional[DataSourceImpl],
        logs: Logs,
    ):
        super().__init__(data_source_impl=data_source_impl, metrics=[])
        self.cte: CTE = cte
        self.dataset_prefix: list[str] = dataset_prefix
        self.dataset_name: str = dataset_name
        self.aggregation_metrics: list[AggregationMetricImpl] = []
        self.data_source_impl: DataSourceImpl = data_source_impl
        self.query_size: int = len(self.build_sql())
        self.logs: Logs = logs

    def can_accept(self, aggregation_metric_impl: AggregationMetricImpl) -> bool:
        sql_expression: SqlExpression = aggregation_metric_impl.sql_expression()
        sql_expression_str: str = self.data_source_impl.sql_dialect.build_expression_sql(sql_expression)
        max_query_length: int = self.data_source_impl.sql_dialect.get_max_sql_statement_length()
        return self.query_size + len(sql_expression_str) < max_query_length

    def append_aggregation_metric(self, aggregation_metric_impl: AggregationMetricImpl) -> None:
        self.aggregation_metrics.append(aggregation_metric_impl)

    def build_sql(self) -> str:
        field_expressions: list[SqlExpression] = self.build_field_expressions()
        select = [
            WITH([self.cte]),
            SELECT(field_expressions),
            FROM(self.cte.alias),
        ]
        self.sql = self.data_source_impl.sql_dialect.build_select_sql(select)
        return self.sql

    def build_field_expressions(self) -> list[SqlExpression]:
        if len(self.aggregation_metrics) == 0:
            # This is to get the initial query length in the constructor
            return [COUNT(STAR())]
        return [aggregation_metric.sql_expression() for aggregation_metric in self.aggregation_metrics]

    def execute(self) -> list[Measurement]:
        sql = self.build_sql()
        try:
            query_result: QueryResult = self.data_source_impl.execute_query(sql)
        except Exception as e:
            if invalid_regex_exc := InvalidRegexException.should_raise(e, sql):
                raise invalid_regex_exc
            raise SodaCoreException(f"Could not execute aggregation query: {e}") from e

        measurements: list[Measurement] = []
        row: tuple = query_result.rows[0]
        for i in range(0, len(self.aggregation_metrics)):
            aggregation_metric_impl: AggregationMetricImpl = self.aggregation_metrics[i]
            measurement_value = aggregation_metric_impl.convert_db_value(row[i])
            measurements.append(
                Measurement(
                    metric_id=aggregation_metric_impl.id,
                    value=measurement_value,
                    metric_name=aggregation_metric_impl.get_short_description(),
                )
            )
        return measurements



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Optional

from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.data_source_results import QueryResult
from soda_core.common.logging_constants import soda_logger
from soda_core.common.metadata_types import ColumnMetadata, SqlDataType
from soda_core.common.utils import format_items
from soda_core.contracts.contract_verification import (
    Check,
    CheckOutcome,
    CheckResult,
    Measurement,
    Threshold,
)
from soda_core.contracts.impl.check_types.schema_check_yaml import SchemaCheckYaml
from soda_core.contracts.impl.contract_verification_impl import (
    CheckImpl,
    CheckParser,
    ColumnImpl,
    ContractImpl,
    MeasurementValues,
    MetricImpl,
    Query,
    ThresholdLevel,
)

logger: logging.Logger = soda_logger


class SchemaCheckParser(CheckParser):
    def get_check_type_names(self) -> list[str]:
        return ["schema"]

    def parse_check(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: SchemaCheckYaml,
    ) -> Optional[CheckImpl]:
        return SchemaCheckImpl(
            contract_impl=contract_impl,
            check_yaml=check_yaml,
        )


@dataclass
class ColumnDataTypeMismatch:
    column: str
    expected_data_type: str
    expected_character_maximum_length: Optional[int]
    actual_data_type: str
    actual_character_maximum_length: Optional[int]

    def get_expected(self) -> str:
        return f"{self.expected_data_type}{self.get_optional_length_str(self.expected_character_maximum_length)}"

    def get_actual(self) -> str:
        return f"{self.actual_data_type}{self.get_optional_length_str(self.actual_character_maximum_length)}"

    @classmethod
    def get_optional_length_str(cls, character_maximum_length: Optional[int]) -> str:
        return f"({character_maximum_length})" if isinstance(character_maximum_length, int) else ""


class SchemaCheckImpl(CheckImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        check_yaml: SchemaCheckYaml,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=None,
            check_yaml=check_yaml,
        )

        logger.info(f"Found {len(contract_impl.column_impls)} columns in contract.")

        self.expected_columns: list[ColumnMetadata] = [
            ColumnMetadata(
                column_name=column_impl.column_yaml.name,
                sql_data_type=(
                    SqlDataType(
                        name=column_impl.column_yaml.data_type,
                        character_maximum_length=column_impl.column_yaml.character_maximum_length,
                    )
                    if column_impl.column_yaml.data_type
                    else None
                ),
            )
            for column_impl in contract_impl.column_impls
        ]
        logger.info(f"Built {len(self.expected_columns)} expected columns from contract.")
        self.allow_extra_columns: bool = bool(check_yaml.allow_extra_columns)
        self.allow_other_column_order: bool = bool(check_yaml.allow_other_column_order)

    def setup_metrics(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: SchemaCheckYaml,
    ):
        self.schema_metric = self._resolve_metric(
            SchemaMetricImpl(
                contract_impl=contract_impl,
            )
        )

        if contract_impl.data_source_impl:
            schema_query: Query = SchemaQuery(
                dataset_prefixes=contract_impl.dataset_prefix,
                dataset_name=contract_impl.dataset_name,
                schema_metric_impl=self.schema_metric,
                data_source_impl=contract_impl.data_source_impl,
            )
            self.queries.append(schema_query)

    def evaluate(self, measurement_values: MeasurementValues) -> CheckResult:
        logger.info("Evaluating schema check")
        outcome: CheckOutcome = CheckOutcome.NOT_EVALUATED

        expected_column_names_not_actual: list[str] = []
        actual_column_names_not_expected: list[str] = []
        column_data_type_mismatches: list[ColumnDataTypeMismatch] = []
        are_columns_out_of_order: bool = False

        logger.info("Getting actual columns from measurement values.")
        # The query to fetch the ColumnMetadata could have failed, in that case no measurement is returned -> actual_columns will be None.
        actual_columns: Optional[list[ColumnMetadata]] = measurement_values.get_value(self.schema_metric)
        if actual_columns is None:
            logger.error("Actual columns are None.")
            logger.error("Setting actual columns to empty list.")
            actual_columns = []

        for column in actual_columns:
            if not isinstance(column, ColumnMetadata):
                logger.error(f"Actual column is not a ColumnMetadata: {column}")
                logger.error("Setting actual columns to empty list.")
                actual_columns = []

        logger.info(f"Found {len(actual_columns)} actual columns.")

        if actual_columns:
            actual_column_names: list[str] = [actual_column.column_name for actual_column in actual_columns]
            actual_column_metadata_by_name: dict[str, ColumnMetadata] = {
                actual_column.column_name: actual_column for actual_column in actual_columns
            }
            expected_column_names: list[str] = [
                expected_column.column_name for expected_column in self.expected_columns
            ]

            logger.info("Starting to check for missing columns.")
            for expected_column in expected_column_names:
                if expected_column not in actual_column_names:
                    expected_column_names_not_actual.append(expected_column)
            logger.info(f"Found {len(expected_column_names_not_actual)} missing columns.")

            if not self.allow_extra_columns:
                logger.info("Starting to check for extra columns.")
                for actual_column_name in actual_column_names:
                    if actual_column_name not in expected_column_names:
                        actual_column_names_not_expected.append(actual_column_name)
                logger.info(f"Found {len(actual_column_names_not_expected)} extra columns.")

            logger.info("Starting to check for data type mismatches.")
            for expected_column in self.expected_columns:
                actual_column_metadata: ColumnMetadata = actual_column_metadata_by_name.get(expected_column.column_name)
                # Assume by default that the data types are the same
                is_same_data_type: bool = True
                try:
                    if actual_column_metadata and expected_column.sql_data_type:
                        is_same_data_type: bool = (
                            self.contract_impl.data_source_impl.sql_dialect.is_same_data_type_for_schema_check(
                                expected=expected_column.sql_data_type,
                                actual=actual_column_metadata.sql_data_type,
                            )
                        )
                except Exception as e:
                    logger.error(
                        f"Error checking for data type mismatch between column {expected_column.column_name} and actual column {actual_column_metadata.column_name}: {e}"
                    )
                    logger.error(f"Skipping column {expected_column.column_name}.")
                    continue
                if not is_same_data_type:
                    column_data_type_mismatches.append(
                        # TODO add numeric_scale, numeric_precision & datetime_precision to the ColumnDataTypeMismatch
                        ColumnDataTypeMismatch(
                            column=expected_column.column_name,
                            expected_data_type=expected_column.sql_data_type.name,
                            expected_character_maximum_length=expected_column.sql_data_type.character_maximum_length,
                            actual_data_type=actual_column_metadata.sql_data_type.name,
                            actual_character_maximum_length=actual_column_metadata.sql_data_type.character_maximum_length,
                        )
                    )
            logger.info(f"Found {len(column_data_type_mismatches)} data type mismatches.")

            if not self.allow_other_column_order:
                logger.info("Starting to check for columns out of order.")
                try:
                    previous_index: int = 0
                    for actual_column_name in actual_column_names:
                        if actual_column_name in expected_column_names:
                            index: int = expected_column_names.index(actual_column_name)
                            if index < previous_index:
                                are_columns_out_of_order = True
                            previous_index = index
                except Exception as e:
                    logger.error(f"Error checking for columns out of order: {e}")
                logger.info(f"Found {are_columns_out_of_order} columns out of order.")

            outcome = (
                CheckOutcome.PASSED
                if (
                    len(expected_column_names_not_actual) == 0
                    and len(actual_column_names_not_expected) == 0
                    and len(column_data_type_mismatches) == 0
                    and not are_columns_out_of_order
                )
                else CheckOutcome.FAILED
            )

        logger.info(f"Schema check outcome: {outcome}")

        return SchemaCheckResult(
            check=self._build_check_info(),
            outcome=outcome,
            expected_columns=self.expected_columns,
            actual_columns=actual_columns,
            expected_column_names_not_actual=expected_column_names_not_actual,
            actual_column_names_not_expected=actual_column_names_not_expected,
            column_data_type_mismatches=column_data_type_mismatches,
            are_columns_out_of_order=are_columns_out_of_order,
        )

    def _build_threshold(self) -> Threshold:
        return Threshold(must_be_less_than_or_equal=0, level=ThresholdLevel.FAIL.value)


class SchemaMetricImpl(MetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
    ):
        super().__init__(contract_impl=contract_impl, metric_type="schema")


class SchemaQuery(Query):
    def __init__(
        self,
        dataset_prefixes: Optional[list[str]],
        dataset_name: str,
        schema_metric_impl: SchemaMetricImpl,
        data_source_impl: Optional[DataSourceImpl],
    ):
        super().__init__(data_source_impl=data_source_impl, metrics=[schema_metric_impl])

        self.sql = data_source_impl.build_columns_metadata_query_str(
            dataset_prefixes=dataset_prefixes, dataset_name=dataset_name
        )

    def execute(self) -> list[Measurement]:
        try:
            query_result: QueryResult = self.data_source_impl.execute_query(self.sql)
        except Exception as e:
            logger.error(msg=f"Could not execute schema query {self.sql}: {e}", exc_info=True)
            return []
        try:
            metadata_columns: list[
                ColumnMetadata
            ] = self.data_source_impl.sql_dialect.build_column_metadatas_from_query_result(query_result)
        except Exception as e:
            logger.error(f"Error building column metadata from query result: {e}")
            return []
        schema_metric_impl: MetricImpl = self.metrics[0]
        return [
            Measurement(metric_id=schema_metric_impl.id, value=metadata_columns, metric_name=schema_metric_impl.type)
        ]


class SchemaCheckResult(CheckResult):
    def __init__(
        self,
        check: Check,
        outcome: CheckOutcome,
        expected_columns: list[ColumnMetadata],
        actual_columns: list[ColumnMetadata],
        expected_column_names_not_actual: list[str],
        actual_column_names_not_expected: list[str],
        column_data_type_mismatches: list[ColumnDataTypeMismatch],
        are_columns_out_of_order: bool,
    ):
        schema_events = sum(
            (
                len(expected_column_names_not_actual),
                len(actual_column_names_not_expected),
                len(column_data_type_mismatches),
            )
        )
        diagnostic_metric_values = {"schema_events_count": schema_events}

        super().__init__(
            check=check,
            outcome=outcome,
            threshold_value=schema_events,
            diagnostic_metric_values=diagnostic_metric_values,
        )

        self.expected_columns: list[ColumnMetadata] = expected_columns
        self.actual_columns: list[ColumnMetadata] = actual_columns
        self.expected_column_names_not_actual: list[str] = expected_column_names_not_actual
        self.actual_column_names_not_expected: list[str] = actual_column_names_not_expected
        self.column_data_type_mismatches: list[ColumnDataTypeMismatch] = column_data_type_mismatches
        self.are_columns_out_of_order: bool = are_columns_out_of_order

    def log_table_row_diagnostics(self, verbose: bool = True) -> str:
        diagnostics = []

        data_delimiter = "\n" if verbose else " "
        category_delimiter = "\n\n" if verbose else "\n"

        if self.actual_column_names_not_expected:
            diagnostics.append(
                f"Present, not expected columns:{data_delimiter}{format_items(self.actual_column_names_not_expected, verbose=verbose)}"
            )

        if self.expected_column_names_not_actual:
            diagnostics.append(
                f"Missing columns:{data_delimiter}{format_items(self.expected_column_names_not_actual, verbose=verbose)}"
            )

        if self.column_data_type_mismatches:
            mismatches = [
                f"{data_type_mismatch.column}: {data_type_mismatch.get_expected()} -> {data_type_mismatch.get_actual()}"
                for data_type_mismatch in self.column_data_type_mismatches
            ]
            diagnostics.append(f"Data type mismatches:{data_delimiter}{format_items(mismatches,verbose=verbose)}")

        if self.are_columns_out_of_order:
            diagnostics.append("There are columns out of order")

        if verbose:
            actual_columns = [
                f"{actual_column.column_name}({actual_column.sql_data_type})" for actual_column in self.actual_columns
            ]
            diagnostics.append(f"Actual schema:{data_delimiter}{format_items(actual_columns, verbose=verbose)}")

        return category_delimiter.join(diagnostics)

    def dataset_does_not_exists(self) -> bool:
        """
        Helper method for testing
        """
        return len(self.actual_columns) == 0

    def column_does_not_exist(self, column_name: str) -> bool:
        """
        Helper method for testing
        """
        return any(actual_column.column_name == column_name for actual_column in self.actual_columns)



from __future__ import annotations

import logging

from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.logging_constants import ExtraKeys, soda_logger
from soda_core.common.sql_dialect import *
from soda_core.contracts.contract_verification import CheckOutcome, CheckResult
from soda_core.contracts.impl.check_types.aggregate_check_yaml import AggregateCheckYaml
from soda_core.contracts.impl.check_types.row_count_check import RowCountMetricImpl
from soda_core.contracts.impl.contract_verification_impl import (
    AggregationMetricImpl,
    CheckImpl,
    CheckParser,
    ColumnImpl,
    ContractImpl,
    MeasurementValues,
    MissingAndValidityCheckImpl,
    ThresholdImpl,
)

logger: logging.Logger = soda_logger


class AggregateCheckParser(CheckParser):
    def get_check_type_names(self) -> list[str]:
        return ["aggregate"]

    def parse_check(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: AggregateCheckYaml,
    ) -> Optional[CheckImpl]:
        return AggregateCheckImpl(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )


class AggregateCheckImpl(MissingAndValidityCheckImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: AggregateCheckYaml,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )
        self.threshold = ThresholdImpl.create(
            threshold_yaml=check_yaml.threshold,
        )

        self.function: str = check_yaml.function.lower()

        # Check if the function is supported by the data source, if data source is available. It is not during `soda contract test`.
        if (
            self.function
            and self.contract_impl.data_source_impl
            and not contract_impl.data_source_impl.sql_dialect.supports_function(self.function)
        ):
            logger.error(
                msg=f"Aggregate function '{check_yaml.function}' is not supported on "
                f"'{contract_impl.data_source_impl.type_name}'",
                extra={ExtraKeys.LOCATION: check_yaml.check_yaml_object.create_location_from_yaml_dict_key("function")},
            )
            self.function = None

    def setup_metrics(
        self,
        contract_impl: ContractImpl,
        column_impl: ColumnImpl,
        check_yaml: AggregateCheckYaml,
    ):
        self.aggregate_metric = self._resolve_metric(
            AggregateFunctionMetricImpl(
                contract_impl=contract_impl, column_impl=column_impl, check_impl=self, function=self.function
            )
        )

        self.check_rows_tested_metric = self._resolve_metric(
            RowCountMetricImpl(contract_impl=contract_impl, check_impl=self)
        )

    def evaluate(self, measurement_values: MeasurementValues) -> CheckResult:
        outcome: CheckOutcome = CheckOutcome.NOT_EVALUATED

        function_value: Optional[float | int] = measurement_values.get_value(self.aggregate_metric)
        check_rows_tested: int = measurement_values.get_value(self.check_rows_tested_metric)

        diagnostic_metric_values: dict[str, float] = {
            "dataset_rows_tested": self.contract_impl.dataset_rows_tested,
            "check_rows_tested": check_rows_tested,
        }

        if isinstance(function_value, Number):
            diagnostic_metric_values[self.function] = function_value

            outcome = self.evaluate_threshold(function_value)

        return CheckResult(
            check=self._build_check_info(),
            outcome=outcome,
            threshold_value=function_value,
            diagnostic_metric_values=diagnostic_metric_values,
        )


class AggregateFunctionMetricImpl(AggregationMetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_impl: MissingAndValidityCheckImpl,
        function: Optional[str],
        column_name: Optional[str] = None,
        data_source_impl: Optional[DataSourceImpl] = None,
        dataset_identifier: Optional[DatasetIdentifier] = None,
    ):
        self.function: Optional[str] = function
        self.column_name = column_impl.column_yaml.name if column_impl else column_name
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            metric_type=check_impl.type,
            check_filter=check_impl.check_yaml.filter,
            missing_and_validity=check_impl.missing_and_validity,
            data_source_impl=data_source_impl,
            dataset_identifier=dataset_identifier,
        )

    def sql_expression(self) -> SqlExpression:
        is_missing = self.missing_and_validity.is_missing_expr(self.column_name)
        is_invalid = self.missing_and_validity.is_invalid_expr(self.column_name)

        filters: list[SqlExpression] = []
        if is_missing or is_invalid:
            filters.append(NOT(AND.optional([is_missing, is_invalid])))
        if self.check_filter:
            filters.append(SqlExpressionStr(self.check_filter))

        arg: SqlExpression | str = self.column_name
        if filters:
            arg = CASE_WHEN(condition=AND(clauses=filters), if_expression=self.column_name, else_expression=None)

        return self.data_source_impl.sql_dialect.get_function_expression(self.function, arg)

    def convert_db_value(self, value) -> any:
        return float(value) if value is not None else None



from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Protocol


class Cursor(Protocol):
    """DB-API 2.0 Cursor protocol.

    This is a subset of the DB-API 2.0 Cursor protocol defined in PEP-0249.
    """

    @property
    def description(self) -> Sequence[tuple[str, Any, Any, Any, Any, Any, Any]]:
        """Gets the description of the columns in the result set.

        Each column is represented as a tuple containing:
            - name
            - type_code
            - display_size
            - internal_size
            - precision
            - scale
            - null_ok

        The first two items (name and type_code) are mandatory, the other five are optional and are set to None if no
        meaningful values can be provided.
        """

    @property
    def rowcount(self) -> int:
        """Gets the number of rows that the last execute*() produced (for DQL statements like SELECT) or affected (for
        DML statements like UPDATE or INSERT).
        """

    def fetchone(self) -> Sequence[Any] | None:
        """Fetches the next row of a query result set, returning a single sequence, or None when no more data is
        available.
        """

    def fetchmany(self, size: int) -> Sequence[Sequence[Any]]:
        """Fetches the next set of rows of a query result, returning a list of sequences. An empty list is returned when
        no more rows are available.
        """

    def fetchall(self) -> Sequence[Sequence[Any]]:
        """Fetches all (remaining) rows of a query result, returning them as a list of sequences."""

    def close(self) -> None:
        """Closes the cursor."""


class QueryResult:
    def __init__(self, rows: list[tuple], columns: tuple[tuple]):
        self.rows: list[tuple] = rows
        self.columns: tuple[tuple] = columns


class QueryResultIterator:
    def __init__(self, cursor: Cursor):
        self._cursor = cursor

    def __iter__(self) -> QueryResultIterator:
        return self

    def __next__(self) -> Sequence[Any]:
        row = self._cursor.fetchone()
        if row is None:
            self._cursor.close()
            raise StopIteration
        return row

    @property
    def row_count(self) -> int:
        """Returns the total number of rows in the result set."""
        return self._cursor.rowcount

    @property
    def columns(self) -> dict[str, Any]:
        """Returns the columns metadata as a dictionary mapping column names to type codes."""
        return {column[0]: column[1] for column in self._cursor.description}


class UpdateResult:
    def __init__(
        self,
        result: object,
    ):
        self.results: object = result



from enum import IntEnum


class ExitCode(IntEnum):
    """Possible exit codes and their meaning.

    See https://docs.soda.io/soda-library/programmatic.html#scan-exit-codes
    """

    OK = 0
    CHECK_FAILURES = 1
    CHECK_WARNINGS = 2
    LOG_ERRORS = 3
    RESULTS_NOT_SENT_TO_CLOUD = 4


OK_CODES = {ExitCode.OK, ExitCode.CHECK_FAILURES}



from __future__ import annotations

from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.sql_dialect import *
from soda_core.contracts.contract_verification import CheckOutcome, CheckResult
from soda_core.contracts.impl.check_types.row_count_check_yaml import RowCountCheckYaml
from soda_core.contracts.impl.contract_verification_impl import (
    AggregationMetricImpl,
    CheckImpl,
    CheckParser,
    ColumnImpl,
    ContractImpl,
    MeasurementValues,
    ThresholdImpl,
    ThresholdType,
)


class RowCountCheckParser(CheckParser):
    def get_check_type_names(self) -> list[str]:
        return ["row_count"]

    def parse_check(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: RowCountCheckYaml,
    ) -> Optional[CheckImpl]:
        return RowCountCheckImpl(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )


class RowCountCheckImpl(CheckImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: RowCountCheckYaml,
    ):
        super().__init__(
            contract_impl=contract_impl,
            column_impl=column_impl,
            check_yaml=check_yaml,
        )

        self.threshold = ThresholdImpl.create(
            threshold_yaml=check_yaml.threshold,
            default_threshold=ThresholdImpl(type=ThresholdType.SINGLE_COMPARATOR, must_be_greater_than=0),
        )
        metric_name: str = ThresholdImpl.get_metric_name(check_yaml.type_name, column_impl=column_impl)
        self.summary = (
            self.threshold.get_assertion_summary(metric_name=metric_name)
            if self.threshold
            else f"{check_yaml.type_name} (invalid threshold)"
        )

    def setup_metrics(
        self,
        contract_impl: ContractImpl,
        column_impl: Optional[ColumnImpl],
        check_yaml: RowCountCheckYaml,
    ):
        self.row_count_metric = self._resolve_metric(RowCountMetricImpl(contract_impl=contract_impl, check_impl=self))

    def evaluate(self, measurement_values: MeasurementValues) -> CheckResult:
        outcome: CheckOutcome = CheckOutcome.NOT_EVALUATED
        row_count: int = measurement_values.get_value(self.row_count_metric)

        outcome = self.evaluate_threshold(row_count)

        diagnostic_metric_values: dict[str, float] = {
            "check_rows_tested": row_count,
            "dataset_rows_tested": self.contract_impl.dataset_rows_tested,
        }

        return CheckResult(
            check=self._build_check_info(),
            outcome=outcome,
            threshold_value=row_count,
            diagnostic_metric_values=diagnostic_metric_values,
        )


class RowCountMetricImpl(AggregationMetricImpl):
    def __init__(
        self,
        contract_impl: ContractImpl,
        check_impl: Optional[CheckImpl] = None,
        filter: Optional[str] = None,
        data_source_impl: Optional[DataSourceImpl] = None,
        dataset_identifier: Optional[DatasetIdentifier] = None,
    ):
        check_filter = filter if filter else check_impl.check_yaml.filter if check_impl else None
        super().__init__(
            contract_impl=contract_impl,
            metric_type="row_count",
            check_filter=check_filter,
            missing_and_validity=None,
            data_source_impl=data_source_impl,
            dataset_identifier=dataset_identifier,
        )

    def sql_expression(self) -> SqlExpression:
        if self.check_filter:
            return SUM(CASE_WHEN(SqlExpressionStr(self.check_filter), LITERAL(1)))
        else:
            return COUNT(STAR())

    def convert_db_value(self, value: any) -> any:
        # Note: expression SUM(CASE WHEN "id" IS NULL THEN 1 ELSE 0 END) gives NULL / None as a result if
        # there are no rows
        return int(value) if value is not None else 0



import re

SODA_CORE_VERSION = "4.0.0.b1"


def clean_soda_core_version() -> str:
    return clean_semver(SODA_CORE_VERSION)


def clean_semver(version_str) -> str:
    match = re.match(r"^(\d+\.\d+\.\d+)", version_str)
    if match:
        return match.group(1)
    else:
        raise ValueError(f"Invalid version format: {version_str}")



from __future__ import annotations

import logging
import os

from dotenv import load_dotenv
from soda_core.common.logging_constants import soda_logger
from soda_core.common.utils import strtobool

logger: logging.Logger = soda_logger


class EnvConfigHelper:
    """
    Helper class to manage environment configuration, feature flags and other config inputs like config files.

    Each attribute is a method for:
    - readibility and consistency
    - to allow for easy mocking in tests
    - to allow overriding of config from different sources (e.g., environment variables, config files) and other logic
    """

    __instance = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
            cls.__instance._initialize()
        return cls.__instance

    @classmethod
    def reset(cls):
        cls.__instance = None

    def _initialize(self):
        logger.debug("Loading environment variables from .env file.")
        load_dotenv(override=True)

    @property
    def soda_core_telemetry_enabled(self) -> bool:
        return strtobool(os.getenv("SODA_CORE_TELEMETRY_ENABLED", "true"))

    @property
    def soda_core_telemetry_local_debug_mode(self) -> bool:
        return strtobool(os.getenv("SODA_CORE_TELEMETRY_LOCAL_DEBUG_MODE", "false"))

    @property
    def soda_core_telemetry_test_mode(self) -> bool:
        return strtobool(os.getenv("SODA_CORE_TELEMETRY_TEST_MODE", "false"))

    @property
    def soda_instruction_id(self) -> str | None:
        return os.getenv("SODA_INSTRUCTION_ID")

    @property
    def is_running_on_agent(self) -> bool:
        # SODA_INSTRUCTION_ID is only set when running in Soda Agent
        return self.soda_instruction_id is not None

    @property
    def soda_scan_definition_type(self) -> str | None:
        return os.getenv("SODA_SCAN_DEFINITION_TYPE")

    @property
    def is_contract_test_scan_definition_type(self) -> bool:
        return self.soda_scan_definition_type == "contractTest"



import logging
import os

from soda_core.common.logging_constants import soda_logger

logger: logging.Logger = soda_logger


def save_to_file(content: str, file_path: str) -> None:
    directory = os.path.dirname(file_path)
    if directory:
        os.makedirs(directory, exist_ok=True)
    with open(file_path, "w") as file:
        file.write(content)

    logger.info(f"Saved file '{file_path}'")



import platform
import types
from typing import Any

from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SimpleSpanProcessor
from opentelemetry.semconv.resource import ResourceAttributes
from soda_core.__version__ import SODA_CORE_VERSION
from soda_core.common.data_source_impl import DataSourceImpl
from soda_core.common.env_config_helper import EnvConfigHelper
from soda_core.contracts.contract_verification import ContractVerificationSessionResult
from soda_core.telemetry.memory_span_exporter import MemorySpanExporter
from soda_core.telemetry.soda_exporter import (
    SodaConsoleSpanExporter,
    SodaOTLPSpanExporter,
)


class SodaTelemetry:
    """Soda Open Telemetry tracing.

    Minimal, anonymous usage telemetry for Soda Core.
    """

    ENDPOINT = "https://collect.soda.io/v1/traces"
    __instance = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
            cls.__instance._initialize()
        return cls.__instance

    def _initialize(self):
        self.soda_config = EnvConfigHelper()

        self.__send = self.soda_config.soda_core_telemetry_enabled

        if self.__send:
            self.__provider = TracerProvider(
                resource=Resource.create(
                    {
                        "os.architecture": platform.architecture(),
                        "python.version": platform.python_version(),
                        "python.implementation": platform.python_implementation(),
                        ResourceAttributes.OS_TYPE: platform.system(),
                        ResourceAttributes.OS_VERSION: platform.version(),
                        "platform": platform.platform(),
                        ResourceAttributes.SERVICE_VERSION: SODA_CORE_VERSION,
                        ResourceAttributes.SERVICE_NAME: "soda",
                        ResourceAttributes.SERVICE_NAMESPACE: "soda-core",
                    }
                )
            )
            if self.soda_config.soda_core_telemetry_test_mode:
                self.__setup_for_test()
            else:
                self.__setup()

    def __setup(self):
        """Set up Open Telemetry processors and exporters for normal use."""
        local_debug_mode = self.soda_config.soda_core_telemetry_local_debug_mode

        if local_debug_mode:
            self.__provider.add_span_processor(BatchSpanProcessor(SodaConsoleSpanExporter()))

        if not local_debug_mode:
            otlp_exporter = SodaOTLPSpanExporter(endpoint=self.ENDPOINT)
            otlp_processor = BatchSpanProcessor(otlp_exporter)
            self.__provider.add_span_processor(otlp_processor)

        trace.set_tracer_provider(self.__provider)

    def __setup_for_test(self):
        self.__provider.add_span_processor(SimpleSpanProcessor(MemorySpanExporter.get_instance()))

        trace.set_tracer_provider(self.__provider)

    def set_attribute(self, key: str, value: str) -> None:
        """Set attribute value in the current span."""
        if self.__send:
            current_span = trace.get_current_span()
            current_span.set_attribute(key, value)

    def set_attributes(self, values: dict[str, str]) -> None:
        """Set attributes the current span."""
        if self.__send:
            current_span = trace.get_current_span()
            current_span.set_attributes(values)

    @staticmethod
    def get_datasource_hash(data_source: DataSourceImpl):
        return data_source.generate_safe_hash()

    def ingest_cli_arguments(self, args: dict[str, Any]) -> None:
        for key, value in args.items():
            self.set_attribute(f"cli__{key}", self.__serialize_arg_value(value))

    def ingest_contract_verification_session_result(
        self, contract_verification_session_result: ContractVerificationSessionResult
    ) -> None:
        self.set_attributes(
            {
                "result__checks_count": contract_verification_session_result.number_of_checks,
                "result__checks_failed_count": contract_verification_session_result.number_of_checks_failed,
                "result__checks_passed_count": contract_verification_session_result.number_of_checks_passed,
            }
        )

    def __serialize_arg_value(self, value):
        if isinstance(value, types.FunctionType):
            return f"{value.__module__}.{value.__qualname__}"
        elif callable(value):
            return f"{type(value).__name__}::{getattr(value, '__name__', repr(value))}"
        elif value is None:
            return ""
        else:
            return value



from __future__ import annotations

from typing import Optional

from soda_core.common.data_source_connection import DataSourceConnection
from soda_core.common.data_source_results import QueryResult
from soda_core.common.sql_ast import *
from soda_core.common.sql_dialect import SqlDialect
from soda_core.common.statements.table_types import (
    FullyQualifiedMaterializedViewName,
    FullyQualifiedObjectName,
    FullyQualifiedTableName,
    FullyQualifiedViewName,
    TableType,
)


class MetadataTablesQuery:
    def __init__(
        self,
        sql_dialect: SqlDialect,
        data_source_connection: DataSourceConnection,
        prefixes: Optional[list[str]] = None,
    ):
        self.sql_dialect = sql_dialect
        self.data_source_connection: DataSourceConnection = data_source_connection
        self.prefixes = prefixes

    def execute(
        self,
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        include_table_name_like_filters: Optional[list[str]] = None,
        exclude_table_name_like_filters: Optional[list[str]] = None,
        types_to_return: Optional[
            list[TableType]
        ] = None,  # To make sure it's backwards compatible with the old behavior, when we use None it should default to [TableType.TABLE]
    ) -> list[FullyQualifiedObjectName]:
        if types_to_return is None:
            types_to_return = [TableType.TABLE]
        select_statement: list = self.build_sql_statement(
            database_name=database_name,
            schema_name=schema_name,
            include_table_name_like_filters=include_table_name_like_filters,
            exclude_table_name_like_filters=exclude_table_name_like_filters,
        )
        sql: str = self.sql_dialect.build_select_sql(select_statement)
        query_result: QueryResult = self.data_source_connection.execute_query(sql)
        return self.get_results(query_result, types_to_return)

    def build_sql_statement(
        self,
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        include_table_name_like_filters: Optional[list[str]] = None,
        exclude_table_name_like_filters: Optional[list[str]] = None,
    ) -> list:
        """
        Builds the full SQL query statement to query table names from the data source metadata.
        """

        join_prefixes = lambda prefixes, schema: [*prefixes, schema] if schema else prefixes
        if self.prefixes is not None:
            prefixes = self.prefixes
        else:
            prefixes = [database_name] if database_name else []
        select: list = [
            FROM(
                self.sql_dialect.table_tables(),
                table_prefix=join_prefixes(prefixes, self.sql_dialect.schema_information_schema()),
            ),
            SELECT(
                [
                    self.sql_dialect.column_table_catalog() or COLUMN(LITERAL(None), field_alias="database_name"),
                    self.sql_dialect.column_table_schema(),
                    self.sql_dialect.column_table_name(),
                    self.sql_dialect.column_table_type() or COLUMN(LITERAL(None), field_alias="table_type"),
                ]
            ),
        ]

        if database_name:
            database_column_name: Optional[str] = self.sql_dialect.column_table_catalog()
            if database_column_name:
                database_name_lower: str = database_name.lower()
                select.append(WHERE(EQ(LOWER(database_column_name), LITERAL(database_name_lower))))

        if schema_name:
            schema_column_name: Optional[str] = self.sql_dialect.column_table_schema()
            if schema_column_name:
                schema_name_lower: str = schema_name.lower()
                select.append(WHERE(EQ(LOWER(schema_column_name), LITERAL(schema_name_lower))))

        table_name_column = self.sql_dialect.column_table_name()

        if include_table_name_like_filters:
            select.append(
                WHERE(
                    OR(
                        [
                            LIKE(LOWER(table_name_column), LITERAL(include_table_name_like_filter.lower()))
                            for include_table_name_like_filter in include_table_name_like_filters
                        ]
                    )
                )
            )

        if exclude_table_name_like_filters:
            for exclude_table_name_like_filter in exclude_table_name_like_filters:
                select.append(
                    WHERE(NOT_LIKE(LOWER(table_name_column), LITERAL(exclude_table_name_like_filter.lower())))
                )

        return select

    def get_results(
        self, query_result: QueryResult, types_to_return: list[TableType]
    ) -> list[FullyQualifiedObjectName]:
        result: list[FullyQualifiedObjectName] = []
        for database_name, schema_name, table_name, table_type in query_result.rows:
            converted_table_type = self.sql_dialect.convert_table_type_to_enum(table_type)
            # If we don't have to return this table table, skip it
            if converted_table_type not in types_to_return:
                continue

            # Create the fully qualified name object
            if converted_table_type == TableType.TABLE:
                result.append(
                    FullyQualifiedTableName(database_name=database_name, schema_name=schema_name, table_name=table_name)
                )
            elif converted_table_type == TableType.VIEW:
                result.append(
                    FullyQualifiedViewName(database_name=database_name, schema_name=schema_name, view_name=table_name)
                )
            elif converted_table_type == TableType.MATERIALIZED_VIEW:
                result.append(
                    FullyQualifiedMaterializedViewName(
                        database_name=database_name, schema_name=schema_name, materialized_view_name=table_name
                    )
                )
            else:
                logger.warning(f"Unexpected table type: {table_type}")

        return result



import logging

soda_logger: logging.Logger = logging.getLogger("soda")
logger: logging.Logger = soda_logger


class Emoticons:
    CROSS_MARK: str = "\u274c"
    WHITE_CHECK_MARK: str = "\u2705"
    WARNING: str = "\u26a0"
    CLOUD: str = "\u2601"
    OK_HAND: str = "\U0001f44c"
    SCROLL: str = "\U0001f4dc"
    FINGERS_CROSSED: str = "\U0001f91e"
    EXPLODING_HEAD: str = "\U0001f92f"
    POLICE_CAR_LIGHT: str = "\U0001f6a8"
    SEE_NO_EVIL: str = "\U0001f648"
    PINCHED_FINGERS: str = "\U0001f90c"
    QUESTION_MARK: str = "\u2753"


class ExtraKeys:
    LOCATION: str = "location"
    DOC: str = "doc"
    EXCEPTION: str = "exception"



import os

from soda_core.cli.exit_codes import ExitCode
from soda_core.common.logging_constants import Emoticons, soda_logger
from soda_core.common.soda_cloud import SodaCloud
from soda_core.contracts.contract_request import (
    RequestStatus,
    fetch_contract_proposal,
    push_contract_proposal,
    transition_contract_request,
)


def handle_fetch_proposal(
    soda_cloud_file_path: str,
    request_number: int,
    output_file_path: str,
    proposal_number: int | None = None,
) -> ExitCode:
    try:
        soda_cloud_client = SodaCloud.from_config(soda_cloud_file_path)
        proposal = fetch_contract_proposal(soda_cloud_client, request_number, proposal_number)
        directory = os.path.dirname(output_file_path)
        if directory:
            os.makedirs(directory, exist_ok=True)
        with open(output_file_path, "w") as file:
            file.write(proposal.contents)
        soda_logger.info(
            f"{Emoticons.WHITE_CHECK_MARK} Successfully fetched proposal {proposal.proposal_number} for request {proposal.request_number} and written it to file {output_file_path}."
        )
        return ExitCode.OK
    except Exception as exc:
        soda_logger.exception(f"An unexpected exception occurred: {exc}")
        return ExitCode.LOG_ERRORS


def handle_push_proposal(
    soda_cloud_file_path: str,
    file_path: str,
    request_number: int,
    message: str | None = None,
) -> ExitCode:
    try:
        soda_cloud_client = SodaCloud.from_config(soda_cloud_file_path)
        with open(file_path) as file:
            contents = file.read()
        proposal = push_contract_proposal(soda_cloud_client, request_number, contents, message)
        comment_msg = f" with comment: {message}" if message else ""
        soda_logger.info(
            f"{Emoticons.WHITE_CHECK_MARK} Successfully pushed proposal {proposal.proposal_number}{comment_msg} for request {request_number} to Soda Cloud."
        )
        return ExitCode.OK
    except Exception as exc:
        soda_logger.exception(f"An unexpected exception occurred: {exc}")
        return ExitCode.LOG_ERRORS


def handle_transition_request(
    soda_cloud_file_path: str,
    request_number: int,
    status: RequestStatus,
) -> ExitCode:
    try:
        soda_cloud_client = SodaCloud.from_config(soda_cloud_file_path)
        request = transition_contract_request(soda_cloud_client, request_number, status)
        soda_logger.info(
            f"{Emoticons.WHITE_CHECK_MARK} Successfully transitioned request {request.request_number} from {request.status.value} to {status.value} on Soda Cloud."
        )
        return ExitCode.OK
    except Exception as exc:
        soda_logger.exception(f"An unexpected exception occurred: {exc}")
        return ExitCode.LOG_ERRORS



import logging
from typing import Dict, Optional, TypeVar

from soda_core.common.logging_constants import soda_logger
from soda_core.common.logs import Logs
from soda_core.common.yaml import ContractYamlSource, SodaCloudYamlSource

T = TypeVar("T", bound="ContractCommandBuilder")


logger: logging.Logger = soda_logger


class ContractCommandBuilder:
    def __init__(self, logs: Optional[Logs] = None):
        self.contract_yaml_sources: list[ContractYamlSource] = []
        self.soda_cloud: Optional["SodaCloud"] = None
        self.soda_cloud_yaml_source: Optional[SodaCloudYamlSource] = None
        self.variables: Optional[Dict[str, str]] = {}
        self.logs: Logs = logs if logs else Logs()
        logger.debug("Publishing contract...")

    def with_contract_yaml_file(self, contract_yaml_file_path: str) -> T:
        if isinstance(contract_yaml_file_path, str):
            logger.debug(f"  ...with contract file path '{contract_yaml_file_path}'")
            self.contract_yaml_sources.append(ContractYamlSource.from_file_path(file_path=contract_yaml_file_path))
        else:
            logger.error(
                f"...ignoring invalid contract yaml file '{contract_yaml_file_path}'. "
                f"Expected string, but was {contract_yaml_file_path.__class__.__name__}."
            )
        return self

    def with_contract_yaml_str(self, contract_yaml_str: str) -> T:
        if isinstance(contract_yaml_str, str):
            logger.debug(f"  ...with contract YAML str [{len(contract_yaml_str)}]")
            self.contract_yaml_sources.append(ContractYamlSource.from_str(yaml_str=contract_yaml_str))
        else:
            logger.error(
                f"...ignoring invalid contract_yaml_str '{contract_yaml_str}'.  "
                f"Expected string, but was {contract_yaml_str.__class__.__name__}"
            )
        return self

    def with_soda_cloud_yaml_file(self, soda_cloud_yaml_file_path: str) -> T:
        if isinstance(soda_cloud_yaml_file_path, str):
            if self.soda_cloud_yaml_source is None:
                logger.debug(f"  ...with soda_cloud_yaml_file_path '{soda_cloud_yaml_file_path}'")
            else:
                logger.debug(
                    f"  ...with soda_cloud_yaml_file_path '{soda_cloud_yaml_file_path}'. "
                    f"Ignoring previously configured soda cloud '{self.soda_cloud_yaml_source}'"
                )
            self.soda_cloud_yaml_source = SodaCloudYamlSource.from_file_path(file_path=soda_cloud_yaml_file_path)
        else:
            logger.error(
                f"...ignoring invalid soda_cloud_yaml_file_path '{soda_cloud_yaml_file_path}'. "
                f"Expected string, but was {soda_cloud_yaml_file_path.__class__.__name__}"
            )
        return self

    def with_soda_cloud_yaml_str(self, soda_cloud_yaml_str: str) -> T:
        if isinstance(soda_cloud_yaml_str, str):
            if self.soda_cloud_yaml_source is None:
                logger.debug(f"  ...with soda_cloud_yaml_str [{len(soda_cloud_yaml_str)}]")
            else:
                logger.debug(
                    f"  ...with soda_cloud_yaml_str '{soda_cloud_yaml_str}'. "
                    f"Ignoring previously configured soda cloud '{self.soda_cloud_yaml_source}'"
                )
            self.soda_cloud_yaml_source = SodaCloudYamlSource.from_str(yaml_str=soda_cloud_yaml_str)
        else:
            logger.error(
                f"...ignoring invalid soda_cloud_yaml_str '{soda_cloud_yaml_str}'. "
                f"Expected string, but was {soda_cloud_yaml_str.__class__.__name__}"
            )
        return self

    def with_soda_cloud(self, soda_cloud: object) -> T:
        logger.debug(f"  ...with provided soda_cloud '{soda_cloud}'")
        self.soda_cloud = soda_cloud
        return self

    def with_variable(self, key: str, value: str) -> T:
        if isinstance(key, str) and isinstance(value, str):
            logger.debug(f"  ...with variable '{key}'")
            self.variables[key] = value
        else:
            logger.error(f"Ignoring invalid variable '{key}'. " f"Expected key str and value string")
        return self

    def with_variables(self, variables: dict[str, str]) -> T:
        if isinstance(variables, dict):
            logger.debug(f"  ...with variables {list(variables.keys())}")
            self.variables.update(variables)
        elif variables is None:
            if isinstance(self.variables, dict) and len(self.variables) > 1:
                logger.debug(f"  ...removing variables {list(self.variables.keys())} because variables set to None")
            self.variables = None
        else:
            logger.error(
                f"Ignoring invalid variables '{variables}'. " f"Expected dict, but was {variables.__class__.__name__}"
            )
        return self
